from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
import pdb

import torch
from torch import nn
from torch.utils.data import DataLoader

from pytrial.utils.check import (
    check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
)
from pytrial.data.site_data import SiteBaseDemographics
from .base import SiteSelectionBase
from .data import TrialSiteSimple, SiteSelectionBaseCollator
from .losses import PolicyGradientLossEnrollment, PolicyGradientLossCombined
from .trainer import SiteSelectTrainer

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
      
class BuildModel(nn.Module):
    def __init__(self, 
        trial_dim,
        site_dim,
        embedding_dim
        ) -> None:
        super().__init__()
        
        self.site_encoder = nn.Linear(site_dim, embedding_dim)
        self.trial_encoder = nn.Linear(trial_dim, embedding_dim)
        self.site_fc = nn.Linear(embedding_dim, embedding_dim)
        self.trial_fc = nn.Linear(embedding_dim, embedding_dim)
        self.score_encoder = nn.Linear(2*embedding_dim, 2*embedding_dim)
        self.fc = nn.Linear(2*embedding_dim, embedding_dim)
        self.output = nn.Linear(embedding_dim, 1)
        
    def forward(self, inputs):
        trial = inputs['trial']
        investigators = inputs['site']
        num_inv = investigators.size(1)
        site_encoding = torch.relu(self.site_fc(torch.relu(self.site_encoder(investigators))))
        trial_encoding = torch.relu(self.trial_fc(torch.relu(self.trial_encoder(trial))))
        trial_encoding = trial_encoding.unsqueeze(1).repeat(1, num_inv, 1)
        network_input = torch.cat((site_encoding, trial_encoding), dim=-1)
        network_input = torch.relu(self.score_encoder(network_input))
        score = self.output(torch.relu(self.fc(network_input))).squeeze(-1)
        return score

class PolicyGradientEntropy(SiteSelectionBase):
    '''
    Implement Policy Gradient Entropy model for selecting clinical trial sites based on possibly missing multi-model site features. [1]_
    
    Parameters
    ----------
    
    trial_dim: list[int]
        Size of the trial representation

    site_dim: int
        Size of the site representation

    embedding_dim: int
        Size of all of the modality and other intermediate embeddings
    
    Notes
    -----
    .. [1] Srinivasa, R. S., Qian, C., Theodorou, B., Spaeder, J., Xiao, C., Glass, L., & Sun, J. (2022). Clinical trial site matching with improved diversity using fair policy learning. arXiv preprint arXiv:2204.06501.

    '''
    def __init__(self, 
        trial_dim=211, 
        site_dim=124, 
        embedding_dim=64, 
        enrollment_only=True,
        K=10,
        lam=1,
        learning_rate=1e-4,
        weight_decay=1e-4,
        batch_size=64,
        epochs=10,
        num_worker=0,
        device='cuda:0',
        experiment_id='test',
        ) -> None:
        super().__init__(experiment_id)
        self.config = {
            'trial_dim':trial_dim,
            'site_dim':site_dim,
            'embedding_dim':embedding_dim,
            'enrollment_only':enrollment_only,
            'K':K,
            'lambda':lam,
            'learning_rate':learning_rate,
            'weight_decay':weight_decay,
            'batch_size':batch_size,
            'epochs':epochs,
            'num_worker':num_worker,
            'device':device,
            'experiment_id':experiment_id,
            }
        self.device = device
        self._build_model()

    def fit(self, train_data):
        '''
        Train model with historical trial-site enrollments.

        Parameters
        ----------
        train_data: TrialSiteSimple
            A `TrialSiteSimple` contains trials, sites, and enrollments.
        '''
        self._input_data_check(train_data)
        self._fit_model(train_data)

    def predict(self, test_data):
        '''
        Make prediction for site selection.
        '''
        selections = []
        self._input_data_check(test_data)
        dataloader = DataLoader(test_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=True,
            shuffle=False,
            collate_fn=SiteSelectionBaseCollator(
                config={
                    'has_demographics':isinstance(test_data.sites, SiteBaseDemographics)
                    }
                ),
            )
        for data in dataloader:
            inputs = self._prepare_input(data)
            scores = self.model(inputs)
            selections += [l[:self.config['K']] for l in scores.argsort(dim=1, descending=True).tolist()]
        return selections

    def save_model(self, output_dir):
        '''
        Save the learned patient-match model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir
        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({'model':self.model}, output_dir=output_dir)

    def load_model(self, checkpoint):
        '''
        Load model and the pre-encoded trial embeddings from the given
        checkpoint dir.

        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.
            If a directory, the only checkpoint file `*.pth.tar` will be loaded.
            If a filepath, will load from this file.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        self.model = state_dict['model']
    
    def get_train_dataloader(self, train_data):
        dataloader = DataLoader(train_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=True,
            shuffle=True,
            collate_fn=SiteSelectionBaseCollator(
                config={
                    'has_demographics': isinstance(train_data.sites, SiteBaseDemographics)
                    }
                ),
            )
        return dataloader

    def _build_model(self):
        self.model = BuildModel(
            trial_dim=self.config['trial_dim'],
            site_dim=self.config['site_dim'],
            embedding_dim=self.config['embedding_dim']
            )
    
    def _build_loss_model(self):
        if self.config['enrollment_only']:
            return PolicyGradientLossEnrollment(self.model, self.config['K'])
        else:
            return PolicyGradientLossCombined(self.model, self.config['K'], self.config['lambda'])     

    def _fit_model(self, train_data):
        train_dataloader = self.get_train_dataloader(train_data)
        loss_model = self._build_loss_model()
        train_objectives = [(train_dataloader, loss_model)]
        trainer = SiteSelectTrainer(
            model=self,
            train_objectives=train_objectives
            )
        trainer.train(**self.config)
        
    def _prepare_input(self, data):
        '''
        Prepare inputs to model.

        Parameters
        ----------
        data: dict[list]
            A batch of trials with their corresponding sites.
        '''
        inputs = {
            'trial': data['trial'].to(self.device),
            'site': data['site'].to(self.device),
            'label': data['label'].to(self.device),
            'eth_label': None if data['eth_label'] is None else data['eth_label'].to(self.device)
            }

        return inputs

    def _input_data_check(self, inputs):
        assert isinstance(inputs, TrialSiteSimple), f'`site_selection` models require input training data in `TrialSiteSimple`, find {type(inputs)} instead.'



'''Implement the data class for trial outcome task'''

import os
import pdb
from pytrial.data.trial_data import TrialOutcomeDatasetBase

class TrialOutcomeDataset(TrialOutcomeDatasetBase):

    def __init__(self, data):
        self.df = data
        self.nct2inc_emb = None
        self.nct2exc_emb = None

    def get_ec_sentence_embedding(self, criteria_column='criteria'):
        '''
        Process the eligibility criteria of each trial,
        get the criterion-level emebddings stored in dict.

        Parameters
        ----------
        criteria_column: str
            The column name of eligibility criteria in the dataframe.
        '''
        if self.inc_ec_embedding is None or self.exc_ec_embedding is None:
            self.criteria_column = criteria_column
            self._process_ec(criteria_column=criteria_column)
            self._collect_cleaned_sentence_set()
            # get the embedding of each criterion
            self._get_ec_emb()

        return {
            "inc_ec_emb": self.inc_ec_embedding, 
            "exc_ec_emb": self.exc_ec_embedding,
            "inc_ec": self.inc_vocab,
            "exc_ec": self.exc_vocab,
            }

    def get_nct_to_ec_emb(self):
        '''
        Get the ec embedding on trial-level, i.e., the average of criterion embeddings.
        Only used when the criterion-level embedding is already obtained by `self.get_ec_sentence_embedding`.
        '''
        if self.inc_ec_embedding is None or self.exc_ec_embedding is None:
            raise ValueError("""The criterion-level embedding is not obtained yet. You should call `self.get_ec_sentence_embedding` first.""")

        if self.nct2inc_emb is None or self.nct2exc_emb is None:
            self.nct2exc_emb = {}
            self.nct2inc_emb = {}
            for idx, row in self.df.iterrows():
                if len(row['inclusion_criteria_index']) > 0:
                    self.nct2inc_emb[row['nctid']] = self.inc_ec_embedding[row['inclusion_criteria_index']].mean(axis=0)
                
                if len(row['exclusion_criteria_index']) > 0:
                    self.nct2exc_emb[row['nctid']] = self.exc_ec_embedding[row['exclusion_criteria_index']].mean(axis=0)

        return {
            "nct2incemb": self.nct2inc_emb,
            "nct2excemb": self.nct2exc_emb,
            }


'''
Implement Trial2Vec model for clinical trial similarity search.
'''

import pdb
import os
import math
from collections import defaultdict
import random

from torch.cuda.amp import autocast
import pandas as pd
import numpy as np
import torch
import transformers
from torch.utils.data import DataLoader
from torch import nn
from transformers import AutoModel, AutoTokenizer
from datasets import load_dataset
from datasets import Dataset
from tqdm import tqdm

from pytrial.utils.trainer import Trainer
from pytrial.utils.check import install_package, make_dir_if_not_exist
from pytrial.utils.check import check_checkpoint_file, check_model_dir, check_model_config_file
from .base import TrialSearchBase
from ..losses import InfoNCELoss
from ..data import TrialSearchCollator, batch_to_device
from ..metrics import precision, recall, ndcg

PRETRAINED_TRIAL2VEC_URL = 'https://storage.googleapis.com/pytrial/trial2vec_pretrained.zip'

class BuildModel(nn.Module):
    config = {}
    def __init__(self,
        model_name,
        emb_dim,
        fields=None,
        ctx_fields=None,
        device=None,
        ) -> None:
        super().__init__()
        self.device = device
        self.base_encoder = AutoModel.from_pretrained(model_name)
        self.global_proj_head = nn.Linear(768, emb_dim, bias=False)
        self.local_proj_head = nn.Linear(768, emb_dim, bias=False)
        self.multihead_att = nn.MultiheadAttention(
            embed_dim=768,
            num_heads=6,
            batch_first=False, # [seq, batch, feature]
        )
        self.config['fields'] = fields
        self.config['ctx_fields'] = ctx_fields

    def forward(self, inputs, return_local_emb=True):
        '''
        (1) Go through each attr and ctx, encode them into embds.
        (2) Go through multihead att over the component embds.
        (3) Apply local and global proj head to get each component embds
        and global trial embds.

        Parameters
        ----------
        inputs: dict[dict]
            A dict of input components after tokenized.
            E.g., inputs['title'] are the tokenized title texts,
            w/ keys like input_ids, attention_mask, etc.

        Returns
        -------
        local_embs: dict[Tensor]
            A dict of all component embds.

        global_embs: Tensor
            The trial-level embeds.
        '''
        local_embs = {}

        fields = self.config['fields']
        ctx_fields = self.config['ctx_fields']

        attr_embs = self._encode_fields(fields, inputs, local_embs) # num_attr, bs, emb_dim

        ctx_embs = self._encode_fields(ctx_fields, inputs, local_embs) # num_ctx, bs, emb_dim

        if len(ctx_embs) > 0:
            # take a fusion of context and attribute embeddings
            temp_embs, att_score = self.multihead_att(ctx_embs, attr_embs, attr_embs) # num_ctx, bs, emb_dim
            temp_embs = torch.mean(temp_embs, 0)
        else:
            # no context embeddings, take directly the average of all local embs
            temp_embs = attr_embs.mean(0) # bs, emb_dim

        # projection and normalize
        global_embs = self.global_proj_head(temp_embs)
        # global_embs = global_embs / global_embs.norm(dim=-1, keepdim=True)

        return_dict = {
            'global_embs':global_embs,
        }

        if return_local_emb:
            return_dict['local_embs'] = local_embs

        return return_dict

    def _encode_fields(self, fields, inputs, local_embs):
        emb_list = []
        for fd in fields:
            input_dict = {}
            for k, v in inputs.items():
                if fd in k:
                    tempkey = k.replace(fd+'_','')
                    input_dict[tempkey] = v
            if len(input_dict) == 0:
                continue
            input_dict = batch_to_device(input_dict, self.device)
            res = self.base_encoder(**input_dict, return_dict=True)
            emb_list.append(res['pooler_output'])
            temp_emb = self.local_proj_head(res['pooler_output'])

            # normalize local embeddings
            # temp_emb = temp_emb / temp_emb.norm(dim=-1, keepdim=True)
            local_embs[fd] = temp_emb
        
        if len(emb_list) > 0:
            embs = torch.stack(emb_list)
            return embs
        else:
            return emb_list

class LocalMatchCollator(TrialSearchCollator):
    def __init__(self,
        bert_name,
        max_seq_length,
        fields,
        ctx_fields,
        tag_field=None,
        is_train=True,
        device='cuda:0',
        ) -> None:
        super().__init__(
            bert_name=bert_name,
            max_seq_length=max_seq_length,
            fields=fields,
            device=device,
            tag_field=tag_field,
        )
        self.fields = fields
        self.ctx_fields = ctx_fields
        self.tag_field = tag_field
        self.is_train = is_train
        self.device = device
        if is_train:
            print('Trigger training for Trial2Vec, will load `nltk` and `textaugment`.')
            install_package('nltk')
            install_package('textaugment')
            import nltk
            from textaugment import EDA
            nltk.download('stopwords')
            nltk.download('omw-1.4')
            nltk.download('wordnet')
            self.eda = EDA()

    def __call__(self, features):
        return_dict = defaultdict(list)
        batch_df = pd.DataFrame(features)
        batch_df.fillna('', inplace=True)

        fields = self.fields
        ctx_fields = self.ctx_fields

        if self.is_train:
            fields = self._random_sample(fields)
            ctx_fields = self._random_sample(ctx_fields)

        return_dict.update(self._batch_tokenize(batch_df=batch_df, fields=fields))
        return_dict.update(self._batch_tokenize(batch_df=batch_df, fields=ctx_fields))

        if self.tag_field is not None:
            return_dict[self.tag_field] = batch_df[self.tag_field].tolist()
        return return_dict

    def _batch_tokenize(self, batch_df, fields):
        return_dict = {}
        for field in fields:

            if self.is_train:
                texts = self._eda_augment(batch_df[field])
            else:
                texts = batch_df[field].tolist()

            tokenized = self.tokenizer(texts, padding=True, truncation=True, max_length=self.max_length, return_tensors='pt')
            return_dict[field] = tokenized
        return return_dict

    def _eda_augment(self, text):
        eda_aug = random.choice([self.eda.synonym_replacement, self.eda.random_swap, self.eda.random_deletion])
        new_text = text.apply(lambda x: eda_aug(x) if len(x)>1 else x)
        new_text = new_text.apply(lambda x: x[0] if isinstance(x,list) else x).tolist()
        return new_text

    def _random_sample(self, fields, n=2):
        fields = np.array(fields).copy()
        if random.random() > 0.5:
            num_select = np.random.poisson(n, 1)[0]
            num_select = np.maximum(num_select, 1)
            sub_fields = fields[:num_select].tolist()
            return sub_fields
        else:
            return fields

class LocalMatchLoss(InfoNCELoss):
    '''
    Match local attribute embeddings and trial-level global embeddings.
    '''
    def __init__(self, model, fields, logit_scale_init_value):
        super().__init__(model=model, logit_scale_init_value=logit_scale_init_value)
        self.fields = fields

    def forward(self, inputs):
        outputs = self.model(inputs)
        embs = outputs['global_embs']
        fields = [f for f in self.fields if f in outputs['local_embs']]
        field = np.random.choice(fields,1)[0]
        local_embs = outputs['local_embs'][field]
        logits_per_trial = self.compute_logits(embs, local_embs)
        logits_per_attr = logits_per_trial.t()
        loss = (self.compute_loss(logits_per_trial) + self.compute_loss(logits_per_attr)) / 2
        return {'loss_value':loss}

class GlobalMatchCollator(LocalMatchCollator):
    def __init__(self,
        bert_name,
        max_seq_length,
        fields,
        ctx_fields,
        tag_field=None,
        is_train=True,
        device='cuda:0',
        ) -> None:
        super().__init__(
            bert_name=bert_name,
            max_seq_length=max_seq_length,
            fields=fields,
            ctx_fields=ctx_fields,
            tag_field=tag_field,
            is_train=is_train,
            device=device,
        )

    def __call__(self, features):
        return_dict = dict()
        batch_df = pd.DataFrame(features)
        batch_df.fillna('', inplace=True)

        fields = self.fields
        ctx_fields = self.ctx_fields

        # process to get anchor
        if self.is_train:
            fields = self._random_sample(fields)
            ctx_fields = self._random_sample(ctx_fields)
            
        if random.random() > 0.5: 
            ctx_fields = [] # only take local embeddings
            ctx_inputs = {}
        else:
            ctx_inputs = self._batch_tokenize(batch_df=batch_df, fields=ctx_fields)

        # anchor is attr+ctx
        return_dict['anchor'] = self._batch_tokenize(batch_df=batch_df, fields=fields)
        return_dict['anchor'].update(ctx_inputs)

        # positive is attr+ctx
        return_dict['pos'] = self._batch_tokenize(batch_df=batch_df, fields=fields)
        return_dict['pos'].update(ctx_inputs)

        # negative is attr'+ctx
        # process to get negative by random shuffling attr field texts
        sub_fields = self._random_sample(fields)
        batch_df[sub_fields] = batch_df[sub_fields].sample(frac=1).reset_index(drop=True)
        return_dict['neg'] = self._batch_tokenize(batch_df=batch_df, fields=fields)
        return_dict['neg'].update(ctx_inputs)

        if self.tag_field is not None:
            return_dict[self.tag_field] = batch_df[self.tag_field].tolist()
        return return_dict

class GlobalMatchLoss(InfoNCELoss):
    '''
    Match positive trials (by replacing trial components).
    '''
    def __init__(self, model, fields, logit_scale_init_value):
        super().__init__(model=model, logit_scale_init_value=logit_scale_init_value)
        self.fields = fields

    def forward(self, inputs):
        outputs_pos = self.model(inputs['pos'], return_local_emb=False)
        outputs_neg = self.model(inputs['neg'], return_local_emb=False)
        outputs_anc = self.model(inputs['anchor'], return_local_emb=False)
        temp_emb = torch.cat([outputs_pos['global_embs'], outputs_neg['global_embs']], dim=0)
        logits_per_trial = self.compute_logits(outputs_anc['global_embs'], temp_emb)
        loss = self.compute_loss(logits_per_trial)
        return {'loss_value': loss}

class Trial2VecTrainer(Trainer):
    '''
    Subclass the original trainer and provide specific evaluation functions.
    '''
    def get_test_dataloader(self, test_data):
        self.test_dataloader = test_data
        return self.test_dataloader

    def prepare_input(self, inputs):
        return self.model._prepare_input(inputs)

    def train_one_iteration(self, 
        max_grad_norm=None,
        warmup_steps=None,
        use_amp=None, 
        scaler=None,
        train_loss_dict=None):
        '''
        Default training one iteration steps, can be subclass can reimplemented.
        '''
        skip_scheduler = False
        num_train_objectives = len(self.train_dataloader)
        for train_idx in range(num_train_objectives):
            data_iterator = self.data_iterators[train_idx]
            loss_model = self.loss_models[train_idx]
            loss_model.zero_grad()
            loss_model.train()
            optimizer = self.optimizers[train_idx]
            scheduler = self.schedulers[train_idx]
            # print(scheduler.get_lr())

            try:
                data = next(data_iterator)
            except StopIteration:
                data_iterator = iter(self.train_dataloader[train_idx])
                self.data_iterators[train_idx] = data_iterator
                data = next(data_iterator)

            # usually pass raw tensors to the target device
            if 'anchor' in data:
                # global match loss
                new_data = {}
                for k,v in data.items():
                    new_data[k] = self.prepare_input(v)
                data = new_data
            else:
                # local match loss
                data = self.prepare_input(data)

            if use_amp:
                with autocast():
                    loss_return = loss_model(data)
                loss_value = loss_return['loss_value']
                scale_before_step = scaler.get_scale()
                scaler.scale(loss_value).backward()
                scaler.unscale_(optimizer)
                torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
                scaler.step(optimizer)
                scaler.update()
                skip_scheduler = scaler.get_scale() != scale_before_step
            else:
                loss_model_return = loss_model(data)
                loss_value = loss_model_return['loss_value']
                loss_value.backward()
                torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
                optimizer.step()

            train_loss_dict[train_idx].append(loss_value.item())

            if use_amp:
                skip_scheduler = scaler.get_scale() != scale_before_step
                if not skip_scheduler:
                    scheduler.step()
            else:
                if warmup_steps > 0:
                    scheduler.step()

    def evaluate(self):
        # encode all trials in test data
        val_doc_embs = self.model.encode(self.model.val_doc_inputs, return_dict=True)
        self.model.update_emb(val_doc_embs)
        return_dict = self.model.evaluate(test_data={'x':self.test_dataloader['x'],'y':self.test_dataloader['y'],})
        return return_dict

class Trial2Vec(TrialSearchBase):
    '''
    Implement the Trial2Vec model for trial document similarity search [1]_.

    Parameters
    ----------
    fields: list[str]
        A list of fields of documents used as the `attribute` fields by Trial2Vec model.

    ctx_fields: list[str]
        A list of fields of documents used as the `context` fields by Trial2Vec model.

    tag_field: str
        The tag indicating trial documents, default to be 'nct_id'.

    bert_name: str (default='emilyalsentzer/Bio_ClinicalBERT')
        The base transformer-based encoder. Please find model names
        from the model hub of transformers (https://huggingface.co/models).

    emb_dim: int, optional (default=768)
        Dimensionality of the embedding vectors.

    logit_scale_init_value: float, optional (default=0.07)
        The logit scale or the temperature.

    max_seq_length: int (default=128)
        The maximum length of input tokens for the base encoder.

    epochs: int, optional (default=10)
        Number of iterations (epochs) over the corpus.

    batch_size: int, optional (default=64)
        Number of samples in each training batch.

    learning_rate: float, optional (default=3e-5)
        The learning rate.

    weight_decay: float, optional (default=1e-4)
        Weight decay applied for regularization.

    warmup_ratio: float (default=0)
        How many steps used for warmup training. If set 0, not warmup.

    evaluation_steps: int (default=10)
        How many iterations while we print the training loss and
        conduct evaluation if evaluator is given.

    num_workers: int, optional (default=0)
        Use these many worker threads to train the model (=faster training with multicore machines).

    device: str or torch.device (default='cuda:0')
        The device to put the model on.

    use_amp: bool (default=False)
        Whether or not use mixed precision training.

    experiment_id: str, optional (default='test')
        The name of current experiment.
    
    Notes
    -----
    .. [1] Wang, Z., & Sun, J. (2022). Trial2Vec: Zero-Shot Clinical Trial Document Similarity Search using Self-Supervision. Findings of EMNLP 2022.
    '''
    trial_embs = {}
    val_doc_inputs = None

    def __init__(self,
        fields=None,
        ctx_fields=None,
        tag_field='nct_id',
        bert_name='emilyalsentzer/Bio_ClinicalBERT',
        emb_dim=128,
        logit_scale_init_value=0.07,
        max_seq_length=128,
        epochs=10,
        batch_size=64,
        learning_rate=2e-5,
        weight_decay=1e-4,
        warmup_ratio=0,
        evaluation_steps=10,
        num_workers=0,
        device='cuda:0',
        use_amp=False,
        experiment_id='test'
        ) -> None:
        super().__init__(experiment_id=experiment_id)
        self.config = {
            'max_seq_length':max_seq_length,
            'logit_scale_init_value':logit_scale_init_value,
            'bert_name':bert_name,
            'emb_dim':emb_dim,
            'epochs':epochs,
            'batch_size':batch_size,
            'learning_rate':learning_rate,
            'weight_decay':weight_decay,
            'evaluation_steps':evaluation_steps,
            'num_workers':num_workers,
            'device':device,
            'use_amp':use_amp,
            'warmup_ratio':warmup_ratio,
            'tag_field':tag_field,
            'fields':fields,
            'ctx_fields':ctx_fields,
        }
        self.tokenizer = AutoTokenizer.from_pretrained(bert_name)
        if num_workers > 1:
            # disable tokenizer parallel when doing data collate parallel
            os.environ['TOKENIZERS_PARALLELISM'] = '0'
        self.use_amp = use_amp
        self.device = device if isinstance(device, str) else 'cuda:0'

        self._build_model()

    def fit(self, train_data, valid_data=None):
        '''Train the trial2vec model to get document embeddings for trial search.

        Parameters
        ----------
        train_data: dict
            train_data = 

            {

            'x': pd.DataFrame,

            'fields': list[str],

            'ctx_fields': list[str],

            'tag': str,

            }

            Training corpus for the model. 

            - `x`: a dataframe of trial documents.
            - `fields`: optional, the fields of documents to use for training as key attributes. If not given, the model uses all fields in `x`.
            - `ctx_fields`: optional, the fields of documents which belong to context components. If not given, the model will only learn from `fields`.
            - `tag`: optional, the field in `x` that serves as unique identifiers. Typically it is the `nct_id` of each trial. If not given, the model takes integer tags.

        valid_data: dict={'x':pd.DataFrame 'y':np.ndarray}.
            Validation data used for identifying the best checkpoint during the training.
            Need to rewrite the function:
            `get_val_dataloader`.

        '''
        self._input_data_check(train_data)
        self._fit(train_data, val_data=valid_data)

    def encode(self,
        inputs,
        batch_size=None,
        num_workers=None,
        return_dict=True,
        verbose=True,
        ):
        '''
        Encode input documents and output the document embeddings.

        Parameters
        ----------
        inputs: dict
        
            inputs = 
            {

            'x': pd.DataFrame,

            'fields': list[str],

            'ctx_fields': list[str],

            'tag': str,

            }

            Share the same input format as the `train_data` in
            `fit` function.
            If `fields`, `ctx_fields`, `tag` are not given,
            will reuse the ones used during training.

        batch_size: int, optional
            The batch size when encoding trials.
        
        num_workers: int, optional
            The number of workers when building the val dataloader.

        return_dict: bool
            If set True, return dict[np.ndarray].
            Else, return np.ndarray with the order same as the input documents.

        verbose: bool
            Whether plot progress bar or not.

        Returns
        -------
        embs: dict[np.ndarray]
            Encoded trial-level embeddings with key (tag) and value (embedding)..
        '''
        self._input_data_check(inputs)

        if batch_size is not None:
            self.config['batch_size'] = batch_size
        
        if num_workers is not None:
            self.config['num_worker'] = num_workers

        # build dataset and dataloader
        dataloader = self.get_val_dataloader(inputs)

        # i = iter(dataloader)
        # res = next(i)

        # go through dataloader and encode trial embds
        emb_list = []
        tag_list = []

        self.model.eval()
        with torch.no_grad():
            for data in tqdm(dataloader, desc='Encoding', disable=not verbose):
                if self.config['tag_field'] is not None:
                    tag = data.pop(self.config['tag_field'])
                    tag_list.append(tag)

                data = self._prepare_input(data)
                outputs = self.model(data)

                temp_emb = outputs['global_embs'].cpu()
                temp_emb = temp_emb / temp_emb.norm(dim=-1, keepdim=True)
                temp_emb = temp_emb.numpy()

                if len(temp_emb.shape) == 1: temp_emb = temp_emb[None]
                emb_list.append(temp_emb)

        embs = np.concatenate(emb_list, 0)

        if len(tag_list) == 0:
            tag_list = list(range(len(embs)))
        else:
            tag_list = sum(tag_list, [])

        if return_dict:
            embs = dict(zip(tag_list, embs))
            return embs
        else:
            return (tag_list, embs)

    def predict(self,
        test_data,
        top_k=10,
        return_df=True,
        skip_pretrained=False,
        ):
        '''
        Predict the top-k relevant for input documents.

        Parameters
        ----------
        test_data: dict
            test_data = 
            
            {

            'x': pd.DataFrame,

            'fields': list[str],

            'ctx_fields': list[str],

            'tag': str,

            }

            Share the same input format as the `train_data` in
            `fit` function.
            If `fields`, `ctx_fields`, `tag` are not given,
            will reuse the ones used during training.

        top_k: int
            Number of retrieved candidates.

        return_df: float
            Whether or not return dataframe for the computed similarity ranking.

            - If set True, return (rank, sim);
            - else, return rank_list=[[(doc1,sim1),(doc2,sim2)], [(doc1,sim1),...]].

        skip_pretrained: bool
            Whether or not skip encoding the trial which has been in the
            self.trial_embs. If set True, will skip encoding the trial, and 
            get the trial embeddings by lookup from self.trial_embs.

        Returns
        -------
        rank: pd.DataFrame
            A dataframe contains the top ranked NCT ids for each.

        sim: pd.DataFrame
            A dataframe contains the corresponding similarities.

        rank_list: list[list[tuple]]
            A list of tuples of top ranked docs and similarities.
        '''
        self._input_data_check(test_data)
        tag_field = self.config['tag_field']

        all_embs = np.stack(self.trial_embs.values())
        all_tags = np.stack(self.trial_embs.keys())

        # skip encoding those already stored in model
        df = test_data['x']

        # debug
        # df['nct_id'].iloc[0] = 'NCT00000001'
        # df['nct_id'].iloc[1] = 'NCT00000002'

        tags, embs = [], []
        if skip_pretrained:
            tags, embs = self._encode_by_lookup(df)
            to_encode_test_trial = df[~df[tag_field].isin(all_tags)]
        else:
            to_encode_test_trial = df

        if len(to_encode_test_trial) > 0:
            # build inputs
            to_encode_test_data = {'x': to_encode_test_trial}
            enc_tags, enc_embs = self.encode(to_encode_test_data, return_dict=False)

            # reorder the trial embs to the same order as the input on the tag_field
            tags += enc_tags

            if len(embs) > 0:
                embs = np.concatenate([embs, enc_embs], 0)
            else:
                embs = enc_embs

            temp_df = pd.DataFrame({tag_field: tags, 'emb': list(embs)})
            temp_df = pd.concat([df.set_index(tag_field), temp_df.set_index(tag_field)], 1).reset_index()
            embs = temp_df['emb'].values
            embs = np.stack(embs, 0)
            tags = temp_df[tag_field].values

        # rank for each test document
        sim = embs.dot(all_embs.T)
        rank = np.argsort(sim, 1)[:,::-1] # flip
        rank = rank[:,1:top_k+1]

        if return_df:
            tag_list, sim_list = [], []
            for i,row in enumerate(rank):
                tag_list.append(all_tags[row])
                sim_list.append(sim[i][row])
            tag_ar = np.array(tag_list)
            sim_ar = np.array(sim_list)
            rank_df = pd.DataFrame(tag_ar, columns=[f'rank_{i}' for i in range(top_k)], index=tags)
            sim_df = pd.DataFrame(sim_ar, columns=[f'rank_{i}' for i in range(top_k)], index=tags)
            return rank_df, sim_df
        else:
            rank_list = []
            for i,row in enumerate(rank):
                tag_ar = all_tags[row]
                sim_ar = sim[i][row]
                rank_list.append(
                    list(zip(tag_ar, sim_ar))
                )
            return rank_list

    def evaluate(self, test_data):
        '''
        Evaluate within the given trial and corresponding candidate trials.

        Parameters
        ----------
        test_data: dict
            test_data =
            
            {
            
            'x': pd.DataFrame,

            'y': pd.DataFrame

            }

            The provided labeled dataset for test trials. Follow the format listed above.

        Returns
        -------
        results: dict[float]
            A dict of metrics and the values.

        Notes
        -----
        x =

        | target_trial | trial1 | trial2 | trial3 |

        | nct01        | nct02  | nct03  | nct04  |

        y =

        | label1 | label2 | label3 |
        
        | 0      | 0      | 1      |
        '''
        test_df = test_data['x']
        label_df = test_data['y']
        ranked_label_list = []
        for idx, row in test_df.iterrows():
            target_trial = row['target_trial']
            if target_trial in self.trial_embs:
                target_emb = self.__getitem__(target_trial)
            else:
                raise ValueError(f'The embeddings of trial {target_trial} are not found.')
            candidate_embs = np.stack([self.__getitem__(tag) for tag in row.values[1:]])
            sim = target_emb[None].dot(candidate_embs.T)[0]
            labels = label_df.iloc[idx].to_numpy()
            if labels.sum() == 0: continue
            ranked_label = labels[np.argsort(sim)[::-1]]
            ranked_label_list.append(ranked_label)
        ranked_label_list = np.array(ranked_label_list)

        return_dict = {}
        for k in [1,2,5]:
            return_dict[f'precision@{k}'] = precision(ranked_label_list, k)
            return_dict[f'recall@{k}'] = recall(ranked_label_list, k)
        return_dict[f'ndcg@{k}'] = ndcg(ranked_label_list, k)
        return return_dict

    def load_model(self, checkpoint):
        '''
        Load model and the pre-encoded trial embeddings from the given
        checkpoint dir.

        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.
            If a directory, the only checkpoint file `*.pth.tar` will be loaded.
            If a filepath, will load from this file.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_model_config(config_filename)
            self.config.update(config)
            self.model.config.update({'fields':config['fields'], 'ctx_fields':config['ctx_fields']})
        self.model.load_state_dict(state_dict['model'])
        self.trial_embs = state_dict['emb']

    def save_model(self, output_dir):
        make_dir_if_not_exist(output_dir)
        self._save_model_config(model_config=self.config, output_dir=output_dir)
        model = self._unwrap_model(self.model)
        self._save_checkpoint(
            {'model':model.state_dict(),'emb':self.trial_embs},
            output_dir=output_dir)

    def get_train_dataloader(self, inputs):
        df = inputs['x']
        dataset_list = self._build_dataset(df=df)
        collator_list = self._build_collator()
        dataloader_list = self._build_dataloader(dataset_list, collator_list)
        return dataloader_list

    def get_val_dataloader(self, inputs):
        df = inputs['x']
        dataset = self._build_val_dataset(df=df)
        collate_fn = self._build_val_collator()
        dataloader = self._build_val_dataloader(dataset=dataset, collate_fn=collate_fn)
        return dataloader

    def update_emb(self, emb_dict):
        '''
        Update trial embeds: add or modify.

        Parameters
        ----------
        emb_dict: dict[np.ndarray]
            The tag and corresponding trial embeddings to updated.
        '''
        self.trial_embs.update(
            emb_dict
        )
    
    def from_pretrained(self, input_dir=None):
        '''
        Download pretrained Trial2Vec model.
        '''
        if input_dir is None or not os.path.exists(input_dir):
            if input_dir is None:
                input_dir = './trial_search/pretrained_trial2vec'

            os.makedirs(input_dir)
            print(f'Download pretrained Trial2Vec model, save to {input_dir}.')
            self._download_pretrained(output_dir=input_dir)
        
        print('Load pretrained Trial2Vec model from', input_dir)
        self.load_model(input_dir)

    def _prepare_input(self, inputs):
        ctx_fields = self.config['ctx_fields']
        fields = self.config['fields']
        output_dict = {}
        for fd in ctx_fields:
            if fd not in inputs:
                continue
            for k,v in inputs[fd].items():
                output_dict[f'{fd}_{k}'] = v

        for fd in fields:
            if fd not in inputs:
                continue
            for k,v in inputs[fd].items():
                output_dict[f'{fd}_{k}'] = v
        return output_dict

    def __getitem__(self, tag):
        return self.trial_embs[tag]

    def _build_model(self):
        model = BuildModel(
            model_name=self.config['bert_name'],
            emb_dim=self.config['emb_dim'],
            ctx_fields=self.config['ctx_fields'],
            fields=self.config['fields'],
            device=self.device,
        )
        self.model = self._wrap_model(model, self.config['device'])

    def _fit(self, train_data, val_data=None):
        # build dataset
        if 'fields' in train_data: self.config['fields']= train_data['fields']
        if 'ctx_fields' in train_data: self.config['ctx_fields'] = train_data['ctx_fields']
        if 'tag' in train_data: self.config['tag_field'] = train_data['tag']
        self.model.config.update({'fields': self.config['fields'], 'ctx_fields':self.config['ctx_fields']})

        if val_data is not None:
            self._build_val_docs(train_data, val_data=val_data)

        loss_models = self._build_loss_model()

        # build train dataloader list
        dataloader_list = self.get_train_dataloader(train_data)

        # kick off training
        train_objectives = list(zip(dataloader_list, loss_models))
        trainer = Trial2VecTrainer(
            model=self,
            train_objectives=train_objectives,
            test_data=val_data,
            test_metric='ndcg@5',
        )

        trainer.train(
            **self.config,
        )

        # encode all training trial docs after training
        self.update_emb(self.encode(train_data, return_dict=True))

    def _build_collator(self, is_train=True):
        collator_list = [
            LocalMatchCollator(
                bert_name=self.config['bert_name'],
                max_seq_length=self.config['max_seq_length'],
                fields=self.config['fields'],
                ctx_fields=self.config['ctx_fields'],
                device=self.config['device'],
                is_train=is_train,
                tag_field=None,
            ),
            GlobalMatchCollator(
                bert_name=self.config['bert_name'],
                max_seq_length=self.config['max_seq_length'],
                fields=self.config['fields'],
                ctx_fields=self.config['ctx_fields'],
                device=self.config['device'],
                is_train=is_train,
                tag_field=None,
            ),
            ]
        return collator_list

    def _build_dataloader(self, dataset_list, collator_list):
        '''Build dataloaders for multiple training
        supervision of Trial2Vec.
        (1) query to the whole trial (w/ the query) search
        (2) replace query w/ others
        '''
        dataloader_list = []
        for i,dataset in enumerate(dataset_list):
            dataloader = DataLoader(
                dataset,
                batch_size=self.config['batch_size'],
                shuffle=True,
                num_workers=self.config['num_workers'],
                collate_fn=collator_list[i],
                pin_memory=True,
            )
            dataloader_list.append(dataloader)
        return dataloader_list

    def _build_dataset(self, df):
        dataset_list = []

        # build local contrastive (match indiv component to global trial embs)
        dataset = Dataset.from_pandas(df)
        dataset.set_transform(self._dataset_transform)
        dataset_list.append(dataset)

        # build global contrastive (discriminate global trial emb when replacing attrs)
        dataset = Dataset.from_pandas(df)
        dataset.set_transform(self._dataset_transform)
        dataset_list.append(dataset)

        return dataset_list

    def _dataset_transform(self, examples):
        return_dict = {}

        if self.config['fields'] is not None:
            for k in self.config['fields']: 
                return_dict[k] = examples[k]
        
        if self.config['ctx_fields'] is not None:
            for k in self.config['ctx_fields']: 
                return_dict[k] = examples[k]

        if self.config['tag_field'] is not None:
            return_dict[self.config['tag_field'] ] = examples[self.config['tag_field'] ]

        if len(return_dict) == 0:
            return_dict = examples

        return return_dict

    def _build_loss_model(self):
        '''
        Build two loss models for training trial2vec from scratch.
        '''
        loss_models = [
            LocalMatchLoss(model=self.model, fields=self.config['fields'], logit_scale_init_value=self.config['logit_scale_init_value']),
            GlobalMatchLoss(model=self.model, fields=self.config['fields'], logit_scale_init_value=self.config['logit_scale_init_value']),
        ]
        return loss_models

    def _val_dataset_transform(self, examples):
        return self._dataset_transform(examples=examples)

    def _build_val_dataset(self, df):
        dataset = Dataset.from_pandas(df)
        dataset.set_transform(self._val_dataset_transform)
        return dataset

    def _build_val_dataloader(self, dataset, collate_fn):
        dataloader = DataLoader(
            dataset,
            batch_size=self.config['batch_size'],
            shuffle=False,
            num_workers=self.config['num_workers'],
            collate_fn=collate_fn,
            pin_memory=True,
        )
        return dataloader

    def _build_val_collator(self):
        return LocalMatchCollator(
            bert_name=self.config['bert_name'],
            max_seq_length=self.config['max_seq_length'],
            fields=self.config['fields'],
            ctx_fields=self.config['ctx_fields'],
            device=self.config['device'],
            is_train=False,
            tag_field=self.config['tag_field'],
        )

    def _build_val_docs(self, train_data, val_data):
        # extract trial docs from val data target trials and candidate trials
        df_tr = train_data['x']
        df_va = val_data['x']
        tag_field = self.config['tag_field']
        tag_field = tag_field if tag_field is not None else 'tag'
        df_va = pd.DataFrame({tag_field: df_va.to_numpy().flatten()}).drop_duplicates().reset_index(drop=True)
        df_va = df_va.merge(df_tr, on=tag_field, how='inner')
        self.val_doc_inputs = {
            'x':df_va,
        }

    def _encode_by_lookup(self, df):
        # get embs and tags by lookup
        tag_field = self.config['tag_field']
        tags, embs = [], []
        for tag in df[tag_field]:
            if tag in self.trial_embs:
                tags.append(tag)
                embs.append(self.trial_embs[tag])
        embs = np.array(embs)
        return tags, embs

    def _input_data_check(self, inputs):
        '''
        Check the training / testing data fits the formats.
        Target to (1) check if inputs valid,
                    if not, give tips about the data problem.

        Parameters
        ----------
        inputs: {
                'x': pd.DataFrame,
                'fields': list[str],
                'ctx_fields': list[str],
                'tag': str,
                }
        '''
        # check overall input format
        assert 'x' in inputs, 'No input trial doc dataframe found in inputs.'
        df = inputs['x']
        if 'fields' in inputs:
            try:
                _ = df[inputs['fields']]
            except:
                raise Exception('Cannot find the specified `fields` in inputs dataframe.')
        if 'ctx_fields' in inputs:
            try:
                _ = df[inputs['ctx_fields']]
            except:
                raise Exception('Cannot find the specified `ctx_fields` in inputs dataframe.')
        if 'tag' in inputs:
            try:
                _ = df[inputs['tag']]
            except:
                raise Exception('Cannot find the specified `tag` in inputs dataframe.')

        # check data type
        try:
            _ = df.applymap(str)
        except:
            raise Exception('Cannot transform the input dataframe to str type, please check the inputs.')

    def _wrap_model(self, model, device):
        if isinstance(device, list):
            model = nn.DataParallel(model, device_ids=device)
            model.to(f'cuda:{model.device_ids[0]}')
        elif device == 'cpu':
            model.to(torch.device('cpu'))
        else:
            model.to(torch.device('cuda'))
        return model

    def _unwrap_model(self, model):
        if isinstance(model, nn.DataParallel):
            return model.module
        else:
            return model

    def _download_pretrained(self, output_dir):
        import wget
        import zipfile
        filename = wget.download(url=PRETRAINED_TRIAL2VEC_URL, out=output_dir)
        zipf = zipfile.ZipFile(filename, 'r')
        zipf.extractall(output_dir)
        zipf.close()


from typing import Any, Callable, Dict, List, NewType, Optional, Tuple, Union
import pdb

import torch
import numpy as np
from torch.utils.data import DataLoader

class Evaluator:
    def __init__(self, model, dataset, collate_fn, device=None):
        self.model = model
        self.dataset = dataset
        self.collate_fn = collate_fn
        self.device = 'cpu' if device is None else device

    def evaluate(self, code_type, ppl_type, eval_batch_size):
        mimic_val_dataset = self.dataset
        mimic_val_collator = self.collate_fn
        mimic_val_collator.set_eval_code_type(code_type)
        mimic_val_collator.set_eval_ppl_type(ppl_type)
        dataloader = DataLoader(mimic_val_dataset,
            batch_size=eval_batch_size,
            num_workers=0,
            drop_last=False,
            collate_fn=mimic_val_collator,
            shuffle=False,
            pin_memory=False)

        ppl_list = []
        for batch in dataloader:
            if batch is not None:
                batch = self._prepare_inputs(batch)
                with torch.no_grad():
                    outputs = self.model(**batch)
                batch_ppl = outputs.perplexity
                batch_ppl = batch_ppl.cpu().flatten().tolist()
                ppl_list.extend(batch_ppl)
        ppl_ar = np.array(ppl_list)
        return np.median(ppl_ar)

    def _prepare_inputs(self, data):
        return type(data)(**{k: self._prepare_input(v) for k, v in data.items()})

    def _prepare_input(self, data: Union[torch.Tensor, Any]) -> Union[torch.Tensor, Any]:
        """
        Prepares one `data` before feeding it to the model, be it a tensor or a nested list/dictionary of tensors.
        """
        if isinstance(data, (tuple, list)):
            return type(data)(self._prepare_input(v) for v in data)
        elif isinstance(data, torch.Tensor):
            kwargs = dict(device=self.device)
            return data.to(**kwargs)
        return data



import abc
import pdb
import os
import json

import torch
from torch import nn
from torch.utils.data import DataLoader
import numpy as np

from pytrial.utils.check import check_model_dir
from pytrial.utils.check import make_dir_if_not_exist
from pytrial.data.patient_data import SequencePatientBase
from pytrial.data.patient_data import SeqPatientCollator
from ..losses import XentLoss, BinaryXentLoss, MSELoss, MultilabelBinaryXentLoss


class SequenceIndivBase(abc.ABC):
    '''Abstract class for all individual outcome predictions
        based on sequential patient data.

    Parameters
    ----------
    experiment_id: str, optional (default = 'test')
        The name of current experiment.
    '''
    _mode_list = ['binary','multiclass','multilabel','regression']
    training=False
    @abc.abstractmethod
    def __init__(self, experiment_id='test', mode=None, output_dim=None):
        check_model_dir(experiment_id)
        self.checkout_dir = os.path.join('./experiments_records', experiment_id,
                                         'checkpoints')
        self.result_dir = os.path.join('./experiments_records', experiment_id,
                                       'results')
        make_dir_if_not_exist(self.checkout_dir)
        make_dir_if_not_exist(self.result_dir)
        self._check_mode_and_output_dim(mode, output_dim)

    @abc.abstractmethod
    def fit(self, train_data, valid_data):
        '''
        Fit function needs to be implemented after subclass.

        Parameters
        ----------
        train_data: Any
            Training data.
        
        valid_data: Any
            Validation data.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, test_data):
        '''
        Prediction function needs to be implemented after subclass.

        Parameters
        ----------
        test_data: Any
            Testing data.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def load_model(self, checkpoint):
        '''
        Load the pretrained model from disk, needs to be implemented after subclass.
        
        Parameters
        ----------
        checkpoint: str
            The path to the checkpoint file.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def save_model(self, output_dir):
        '''
        Save the model to disk, needs to be implemented after subclass.

        Parameters
        ----------
        output_dir: str
            The path to the output directory.
        '''
        raise NotImplementedError

    def train(self, mode=True):
        '''
        Swith the model to the `training` mode. Work samely as `model.train()` in pytorch.

        Parameters
        ----------
        mode: bool, optional (default = True)
            If True, switch to the `training` mode.
        '''
        self.training = mode
        self.model.train()
        return self
    
    def eval(self, mode=False):
        '''
        Swith the model to the `validation` mode. Work samely as `model.eval()` in pytorch.

        Parameters
        ----------
        mode: bool, optional (default = False)
            If False, switch to the `validation` mode.
        '''
        self.training = mode
        self.model.eval()
        return self

    def get_test_dataloader(self, test_data):
        dataloader = DataLoader(test_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=True,
            shuffle=False,
            collate_fn=SeqPatientCollator(
                config={
                    'visit_mode':test_data.metadata['visit']['mode'],
                    'label_mode':test_data.metadata['label']['mode'],
                    }
                ),
            )
        return dataloader

    def get_train_dataloader(self, train_data):
        dataloader = DataLoader(train_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=False,
            shuffle=True,
            collate_fn=SeqPatientCollator(
                config={
                    'visit_mode':train_data.metadata['visit']['mode'],
                    'label_mode':train_data.metadata['label']['mode'],
                    }
                ),
            )
        return dataloader

    def _input_data_check(self, inputs):
        assert isinstance(inputs, SequencePatientBase), 'Wrong input type.'

    def _check_mode_and_output_dim(self, mode, output_dim):
        mode = mode.lower()
        assert mode in self._mode_list, f'Input mode `{mode}` does not belong to the supported mode list {self._mode_list}.'
        if output_dim is None:
            if mode not in ['binary','regression']:
                raise ValueError('`output_dim` should be given when `mode` is not `binary` or `regression`.')
            else:
                output_dim = 1
        self.mode = mode
        self.output_dim = output_dim
    
    def _save_config(self, config, output_dir):
        if output_dir is None:
            output_dir = self.checkout_dir

        temp_path = os.path.join(output_dir, 'config.json')
        if os.path.exists(temp_path):
            os.remove(temp_path)
        with open(temp_path, 'w', encoding='utf-8') as f:
            f.write(
                json.dumps(config, indent=4)
            )

    def _load_config(self, checkpoint=None):
        '''
        Load model config from the given directory.
        Parameters
        ----------
        checkpoint: str
            The given filepath (e.g., ./checkpoint/config.json)
            to load the model config.
        '''
        if checkpoint is None:
            temp_path = os.path.join(checkpoint, 'config.json')
        else:
            temp_path = checkpoint

        assert os.path.exists(temp_path), 'Cannot find `config.json` under {}'.format(self.checkout_dir)
        with open(temp_path, 'r') as f:
            config = json.load(f)
        return config

    def _save_checkpoint(self, 
        state_dict,
        epoch_id=0,
        is_best=False,
        output_dir=None, 
        filename='checkpoint.pth.tar'
        ):
        if output_dir is None:
            output_dir = self.checkout_dir

        if epoch_id < 1:
            filepath = os.path.join(output_dir, 'latest.' + filename)
        elif is_best:
            filepath = os.path.join(output_dir, 'best.' + filename)
        else:
            filepath = os.path.join(self.checkout_dir,
                                    str(epoch_id) + '.' + filename)
        torch.save(state_dict, filepath)

    @torch.no_grad()
    def _predict_on_dataloader(self, test_dataloader):
        pred_list, label_list = [], []
        for batch in test_dataloader:
            inputs = self._prepare_input(batch)
            logits = self.model(inputs)
            pred_list.append(logits)
            if 'y' in batch: label_list.append(batch.pop('y'))
        pred = torch.cat(pred_list, dim=0)
        if self.config['mode'] in ['binary','multilabel']:
            pred = torch.sigmoid(pred)
        if self.config['mode'] == 'multiclass':
            pred = torch.softmax(pred, dim=1)
        pred = pred.cpu().numpy()
        label = torch.cat(label_list) if len(label_list) > 0 else None
        return {'pred':pred,'label':label}

    def _build_loss_model(self):
        mode = self.config['mode']
        if mode == 'binary':
            return [BinaryXentLoss(self.model)]

        if mode == 'multiclass':
            return [XentLoss(self.model)]

        if mode == 'multilabel':
            return [MultilabelBinaryXentLoss(self.model)]

        if mode == 'regression':
            return [MSELoss(self.model)]

    def _prepare_input(self, data):
        '''
        Prepare inputs for sequential patient record predictive models.

        Parameters
        ----------
        data: dict[list]
            A batch of patient records.

        '''
        visits = data['v']
        feature = data['x']

        if not isinstance(feature, torch.Tensor): feature = torch.tensor(feature)
        feature = feature.to(self.device)

        inputs = {
            'v':{},
            'v_lengths':[],
            'x':feature, # baseline feature
            }
        
        if self.config['max_visit'] is None:
            v_lengths = [len(visits[self.config['orders'][0]][idx]) for idx in range(len(visits[self.config['orders'][0]]))]
        else:
            max_visit = self.config['max_visit']
            v_lengths = [len(visits[self.config['orders'][0]][idx][:max_visit]) for idx in range(len(visits[self.config['orders'][0]]))]

        inputs['v_lengths'] = v_lengths

        v = torch.zeros(len(v_lengths), max(v_lengths), self.config['total_vocab_size'])
        for idx in range(len(v_lengths)):
            v[idx,:v_lengths[idx]] = torch.tensor(self._translate_dense_visits_to_sparse({k: visits[k][idx][:self.config['max_visit']] for k in visits}))
        v = v.to(self.device)
        inputs['v'] = v

        if 'y' in data: # target labels
            target = data['y']
            if not isinstance(target, torch.Tensor): target = torch.tensor(target)
            target = target.to(self.device)
            inputs['y'] = target
        return inputs

    def _translate_dense_visits_to_sparse(self, visits):
        total_vocab_size = sum(self.config['vocab_size'])
        num_visits = len(visits[self.config['orders'][0]])
        outputs = np.zeros((num_visits, total_vocab_size))

        for i, o in enumerate(self.config['orders']):
            for j in range(num_visits):
                raw = visits[o][j]
                if isinstance(raw, torch.Tensor): raw = raw.detach().cpu().numpy()
                if i > 0: 
                    voc_size = sum(self.config['vocab_size'][:i-1])
                    if isinstance(raw, list):
                        raw = [r + voc_size for r in raw]
                    else:
                        raw += voc_size
                outputs[j, raw] = 1

        return outputs



class InputEventEmbedding(nn.Module):
    def __init__(self, orders, vocab_size, emb_size, padding_idx) -> None:
        super().__init__()
        # build input embeddings
        emb_dict = {}
        for i, order in enumerate(orders):
            emb_dict[order] = nn.Embedding(vocab_size[i], embedding_dim=emb_size, padding_idx=padding_idx)
        self.embeddings = nn.ModuleDict(emb_dict)
    
    def forward(self, inputs):
        emb_list = []
        for k, v in inputs['v'].items():
            emb = self.embeddings[k](v)
            emb_list.append(emb)
        embs = torch.cat(emb_list, 1)
        return embs

class RNNModel(nn.Module):
    RNN_TYPE = {
        'rnn':nn.RNN,
        'lstm':nn.LSTM,
        'gru':nn.GRU,
    }
    def __init__(self, 
        rnn_type,
        emb_size,
        num_layer,
        bidirectional,
        ) -> None:
        super().__init__()
        self.model = self.RNN_TYPE[rnn_type](
            input_size=emb_size,
            hidden_size=emb_size,
            num_layers=num_layer,
            bidirectional=bidirectional,
            batch_first=True,
        )
        self.bidirectional = bidirectional
    
    def forward(self, x):
        outputs = self.model(x)[0]
        return outputs


'''
Implement CTGAN model for tabular simulation
prediction in clinical trials.
'''
import os
import warnings
import joblib

from .base import TabularSimulationBase
from .copula_gan import CTGANSynthesizer

from pytrial.data.patient_data import TabularPatientBase
from pytrial.utils.check import check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist

warnings.filterwarnings('ignore')


class BuildModel:
    def __new__(self, config) -> CTGANSynthesizer:
        model = CTGANSynthesizer(
            embedding_dim=config['embedding_dim'],
            generator_dim=config['generator_dim'],
            discriminator_dim=config['discriminator_dim'],
            generator_lr=config['generator_lr'],
            generator_decay=config['generator_decay'],
            discriminator_lr=config['discriminator_lr'],
            discriminator_decay=config['discriminator_decay'],
            batch_size=config['batch_size'],
            discriminator_steps=config['discriminator_steps'],
            log_frequency=config['log_frequency'],
            verbose=config['verbose'],
            epochs=config['epochs'],
            pac=config['pac'],
            cuda=config['cuda'],
            )

        return model


class CTGAN(TabularSimulationBase):
    '''
    Implement CTGAN model for patient level tabular data generation [1]_.

    Parameters
    ----------
    embedding_dim: int
        Size of the random sample passed to the Generator. Defaults to 128.

    generator_dim: tuple or list of ints
        Size of the output samples for each one of the Residuals. A Residual Layer
        will be created for each one of the values provided. Defaults to (256, 256).
    
    discriminator_dim: tuple or list of ints
        Size of the output samples for each one of the Discriminator Layers. A Linear Layer
        will be created for each one of the values provided. Defaults to (256, 256).
    
    generator_lr: float
        Learning rate for the generator. Defaults to 2e-4.
    
    generator_decay: float
        Generator weight decay for the Adam Optimizer. Defaults to 1e-6.
    
    discriminator_lr: float
        Learning rate for the discriminator. Defaults to 2e-4.
    
    discriminator_decay: float
        Discriminator weight decay for the Adam Optimizer. Defaults to 1e-6.
    
    batch_size: int
        Number of data samples to process in each step.
    
    discriminator_steps: int
        Number of discriminator updates to do for each generator update.
        From the WGAN paper: https://arxiv.org/abs/1701.07875. WGAN paper
        default is 5. Default used is 1 to match original CTGAN implementation.
    
    log_frequency: bool
        Whether to use log frequency of categorical levels in conditional
        sampling. Defaults to ``True``.
    
    verbose: bool
        Whether to have print statements for progress results. Defaults to ``True``.
    
    epochs: int
        Number of training epochs. Defaults to 300.
    
    pac: int
        Number of samples to group together when applying the discriminator.
        Defaults to 10.
    
    cuda: bool or str
        If ``True``, use CUDA. If a ``str``, use the indicated device.
        If ``False``, do not use cuda at all.

    experiment_id: str, optional (default='trial_simulation.tabular.ctgan')
        The name of current experiment. Decide the saved model checkpoint name.

    Notes
    -----
    .. [1] Xu, L., Skoularidou, M., Cuesta-Infante, A., & Veeramachaneni, K. (2019). Modeling tabular data using conditional gan. Advances in Neural Information Processing Systems, 32.
    '''
    def __init__(
            self,
            embedding_dim=128,
            generator_dim=(256, 256),
            discriminator_dim=(256, 256),
            generator_lr=2e-4,
            generator_decay=1e-6,
            discriminator_lr=2e-4,
            discriminator_decay=1e-6,
            batch_size=500,
            discriminator_steps=1,
            log_frequency=True,
            verbose=True,
            epochs=50,
            pac=10,
            cuda=False, # can be set to "True" if applicable
            experiment_id='trial_simulation.tabular.ctgan',
    ) -> None:
        super().__init__(experiment_id=experiment_id)
        self.config = {
            'embedding_dim' : embedding_dim,
            'generator_dim' : generator_dim,
            'discriminator_dim' : discriminator_dim,
            'generator_lr' : generator_lr,
            'generator_decay' : generator_decay,
            'discriminator_lr' : discriminator_lr,
            'discriminator_decay' : discriminator_decay,
            'batch_size' : batch_size,
            'discriminator_steps' : discriminator_steps,
            'log_frequency' : log_frequency,
            'verbose' : verbose,
            'epochs' : epochs,
            'pac' : pac,
            'cuda' : cuda,
            'experiment_id': experiment_id,
            'model_name': 'ct_gan',
        }
        self._save_config(self.config)

    def fit(self, train_data):
        '''
        Train CTGAN model to simulate tabular patient data.
        
        Parameters
        ----------
        train_data: TabularPatientBase
            The training data.
        '''
        self._input_data_check(train_data)
        self._build_model()
        if isinstance(train_data, TabularPatientBase):  # transform=True
            self.metadata = train_data.metadata
            self.raw_dataset = train_data
        if isinstance(train_data, dict):
            tabular_patient_base_dataset = TabularPatientBase(train_data, transform=True)
            self.metadata = tabular_patient_base_dataset.metadata
            self.raw_dataset = tabular_patient_base_dataset
        dataset = self.raw_dataset.reverse_transform()  # transform back

        categoricals = []
        fields_names = self.metadata['sdtypes']
        for field in dataset.columns:
            field_name = field.replace('.value', '')
            if field_name in fields_names:
                meta = fields_names[field_name]
                if meta == 'categorical':
                    categoricals.append(field)
        self._fit_model(dataset, categoricals)

    def predict(self, n=200):
        '''
        simulate a new tabular data with number_of_predictions.

        Parameters
        ----------
        n: int
            number of synthetic records going to generate.

        Returns
        -------
        ypred: dataset, same as the input dataset
            A new tabular data simulated by the model
        '''
        ypred = self.model.sample(n)  # build df
        return ypred  # output: dataset, same as the input dataset, don't need to transform back

    def save_model(self, output_dir=None):
        '''
        Save the learned CTGAN model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is None:
            output_dir = self.checkout_dir

        make_dir_if_not_exist(output_dir)

        self._save_config(self.config, output_dir=output_dir)
        ckpt_path = os.path.join(output_dir, 'ctgan.model')
        joblib.dump(self.model, ckpt_path)

    def load_model(self, checkpoint=None):
        '''
        Save the learned CTGAN model to the disk.

        Parameters
        ----------
        checkpoint: str or None
            If a directory, the only checkpoint file `.model` will be loaded.
            If a filepath, will load from this file;
            If None, will load from `self.checkout_dir`.
        '''
        if checkpoint is None:
            checkpoint = self.checkout_dir

        checkpoint_filename = check_checkpoint_file(checkpoint, suffix='model')
        config_filename = check_model_config_file(checkpoint)
        self.model = joblib.load(checkpoint_filename)
        self.config = self._load_config(config_filename)

    def _build_model(self):
        self.model = BuildModel(self.config)

    def _fit_model(self, data, discrete_columns):
        self.model.fit(data, discrete_columns=discrete_columns)



import os
import setuptools

this_directory = os.path.abspath(os.path.dirname(__file__))

with open("README.md", "r") as f:
    long_description = f.read()

# read the contents of requirements.txt
with open(os.path.join(this_directory, 'requirements.txt'),
          encoding='utf-8') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name = 'PromptEHR',
    version = '0.0.6',
    author = 'Zifeng Wang',
    author_email = 'zifengw2@illinois.edu',
    description = 'Sequence patient electronic healthcare record generation with large language models (LLMs) as the neural database.',
    url = 'https://github.com/RyanWangZf/PromptEHR',
    keywords=['synthetic data','healthcare','EHR','deep learning','AI'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(exclude=['test']),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
)


'''
Implement SPOT model for clinical trial outcome prediction.
'''
import pdb
import os
import pickle
import math
import json
import wget
from collections import defaultdict

import numpy as np
import torch
from torch import nn
import torch.nn.init as nn_init
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader
from torch.cuda.amp import autocast

from pytrial.utils.trainer import Trainer
from .base import TrialOutcomeBase
from .model_utils.topic_discovery import TrialTopicDiscovery
from .model_utils.molecule_encode import MPNN, ADMET
from .model_utils.icdcode_encode import GRAM
from .model_utils.module import Highway
from .model_utils.data_utils import collect_trials_from_topic
from .evaluation import evaluate



__all__ = ['SPOT']

class constants:
    YEAR_RANGE = [1950, 2050]
    ICDCODE_ANCESTOR_FILENAME = './demo_data/icdcode_data/icdcode2ancestor_dict.pkl'
    ICDCODE_LIST_FILENAME = './demo_data/icdcode_data/raw_data.csv'
    RAW_DATA_FILENAME = './demo_data/icdcode_data/raw_data.csv'
    INC_EMB_FILENAME = './demo_data/criteria_data/nct2inc_embedding.pkl'
    EXC_EMB_FILENAME = './demo_data/criteria_data/nct2exc_embedding.pkl'
    WEIGHT_NAME = 'pytorch_model.bin'
    CONFIG_NAME = 'config.json'
    TOPIC_DISCOVER_NAME = 'topic_discover.pkl'
    MODEL_OBJECT_NAME = 'spot_model.pkl'
    SEQDATA_NAME = 'seqtrial_data.pkl'
    TRAINING_PARAMETERS = ['batch_size','learning_rate','epochs','weight_decay','warmup_ratio','evaluation_steps','seed']
    # used for downloading pretrained embeddings / preprocessed data
    BENCHMARK_DATA_URL = 'https://storage.googleapis.com/pytrial/HINT-benchmark-data/hint_benchmark_dataset_w_date.zip'
    ICDCODE_ANCESTOR_URL = 'https://storage.googleapis.com/pytrial/HINT-benchmark-data/icdcode2ancestor_dict.pkl'
    INC_EMB_URL = 'https://storage.googleapis.com/pytrial/HINT-benchmark-data/nct2inc_embedding.pkl'
    EXC_EMB_URL = 'https://storage.googleapis.com/pytrial/HINT-benchmark-data/nct2exc_embedding.pkl'
    ADMET_MODEL_URL = 'https://storage.googleapis.com/pytrial/HINT-benchmark-data/admet_model_state_dict.ckpt'

def check_nan(inputs):
    # check nan for debugging
    if torch.isnan(inputs).any():
        pdb.set_trace()

class SPOTCollator:
    def __call__(self, inputs):
        '''
        Parameters
        ----------
        inputs: List[Topic]
            A list of `Topic` objects.
        '''
        return_data = defaultdict(list)
        return_data['topic'] = inputs
        # create mask list, (bs, n_ts, max_n_trial)
        for input in inputs:
            mask = []
            for trial_step in input.trials:
                mask.append(torch.tensor([1] * len(trial_step)))
            # pad mask
            mask = pad_sequence(mask, batch_first=True, padding_value=0)
            return_data['mask'].append(mask)
        return return_data

class MoleculeEncoder(nn.Module):
    '''
    Load pretrained MPNN and ADMET models.
    '''
    def __init__(self, input_dir=None, device='cuda:0'):
        super(MoleculeEncoder, self).__init__()
        if input_dir is None:
            input_dir = './pretrained_model/molecule_encoder'
        if not os.path.exists(input_dir):
            os.makedirs(input_dir)
            # download pretrained model if not exist from the url ADMET_MODEL_URL
            self._download_pretrained(input_dir)
        
        # load pretrained admet model and mpnn model
        state_dict = torch.load(os.path.join(input_dir, 'admet_model_state_dict.ckpt'))
        self.mpnn = MPNN(mpnn_hidden_size=50, mpnn_depth=3, device=device)
        self.admet = ADMET(molecule_encoder=self.mpnn,
            highway_num=2,
            device=device,
            epoch=None,
            lr=None,
            weight_decay=None,
            save_name=None,
            )
        self.admet.load_state_dict(state_dict)

        # freeze admet model
        # for param in self.admet.parameters():
        #     param.requires_grad = False
             
        self.pk_fc = nn.Linear(50 * 5, 50)
        self.pk_highway = Highway(50, 2)
    
    def forward(self, smiles):
        # get molecule features
        res = self.mpnn.forward_smiles_lst(smiles) # (n_molecule, 50)
        # if w/o admet, return mpnn features
        return res

    def _download_pretrained(self, input_dir):
        wget.download(constants.ADMET_MODEL_URL, input_dir)

class DiseaseEncoder(nn.Module):
    def __init__(self, device) -> None:
        super().__init__()
        self.gram = GRAM(
            embedding_dim=50, 
            icdcode2ancestor=None, 
            device=device,
        )
        self.device = device
    
    def forward(self, icdcode):
        # in one sample, icdcode is a list of list of icdcode belonging to the same trial
        res = self.gram.forward_code_lst2(icdcode)
        return res

class ProtocolEncoder(nn.Module):
    def __init__(self, output_dim=50) -> None:
        super().__init__()
        self.nct2incemb = nn.ParameterDict({})
        self.nct2excemb = nn.ParameterDict({})
        self.fc = nn.Linear(768, output_dim)
        self.activation = nn.ReLU()

    def forward(self, nctid):
        '''
        Get the pre-encoded protocol embeddings.
        '''
        device = self.fc.weight.device
        # say if nctid in the dict, return the embedding
        # else return an all zero embedding
        if nctid in self.nct2incemb:
            inc_emb = self.nct2incemb[nctid][None].to(device)
        else:
            inc_emb = torch.zeros(1, 768).to(device)

        if nctid in self.nct2excemb:
            exc_emb = self.nct2excemb[nctid][None].to(device)
        else:
            exc_emb = torch.zeros(1, 768).to(device)

        emb = torch.cat([inc_emb, exc_emb], dim=0) # [2, 768]
        emb = self.fc(emb)
        emb = self.activation(emb) # [2, 50]
        return emb
    
    def update_criteria_emb(self, inc_emb_dict=None, exc_emb_dict=None):
        '''
        Update the nct2incemb and nct2excemb dict.
        '''
        if inc_emb_dict is not None:
            self.nct2incemb.update(inc_emb_dict)
            self._disable_grad(self.nct2incemb)
                
        if exc_emb_dict is not None:
            self.nct2excemb.update(exc_emb_dict)
            self._disable_grad(self.nct2excemb)

    def _disable_grad(self, nct2incemb):
        for k,v in nct2incemb.items():
            nct2incemb[k] = v.requires_grad_(False)
        return nct2incemb

class FeatureEncoder(nn.Module):
    def __init__(self, input_dim, output_dim=50):
        super().__init__()
        self.fc = nn.Linear(input_dim, output_dim)
        nn_init.uniform_(self.fc.weight, a=-1/math.sqrt(output_dim),b=1/math.sqrt(output_dim))

        # assign each year an embedding, drawback is that some years are not available in the training data
        self.embed = nn.Embedding(150, output_dim)
        nn_init.uniform_(self.embed.weight, a=-1/math.sqrt(output_dim),b=1/math.sqrt(output_dim))

    def forward(self, x):
        '''
        Currently only encode year.
        Args:
            x: int
        '''
        # assign each year an embedding, drawback is that some years are not available in the training data
        # x = torch.tensor(x-1900).long().to(self.embed.weight.device)
        # x = self.embed(x)[None]

        # x maps to [0, 1], assume x is in [1950, 2050]
        start, end = constants.YEAR_RANGE
        x = torch.tensor((x - start)/(end-start), device=self.fc.weight.device).unsqueeze(0)
        x = self.fc(x)[None]
        return x

class TopicEncoder(nn.Module):
    def __init__(self, num_topics, output_dim=50):
        super().__init__()
        self.embed = nn.Embedding(num_topics, output_dim)
        nn_init.uniform_(self.embed.weight, a=-1/math.sqrt(output_dim),b=1/math.sqrt(output_dim))

    def forward(self, x):
        '''
        Currently only encode a single topic id in str or int.
        '''
        if isinstance(x, str):
            x = int(x)
        x = torch.tensor(x, device=self.embed.weight.device)
        x = self.embed(x)[None]
        return x


class SPOTTrainer(Trainer):
    '''
    Make supervised sequential modeling (no meta learning).
    '''
    def prepare_input(self, data):
        return self.model.prepare_input(data, return_label=True)
    
    def evaluate(self):
        res = self.model._predict_on_dataloader(self.test_dataloader)
        pred, label, nctid = res['pred'], res['label'], res['nctid']
        bool_idx = np.isin(nctid, self.test_data.new_trials[-1]) # 0: valids, 1: testids

        eval_label = label[bool_idx]
        eval_pred = pred[bool_idx]

        eval_res = evaluate(eval_pred, eval_label, find_best_threshold=True)
        return eval_res
    
    def get_test_dataloader(self, test_data):
        '''
        Build a dataloader for testing a spot model.
        Parameters
        ----------
        test_data: SequenceTrial
            A dataset consisting of multiple topics of clinical trials. Each topic is a list of clinical trials under this topic ordered with their timestamps.
        '''
        self.test_data = test_data
        loader = self.model.get_test_dataloader(test_data)
        return loader

    def train_one_iteration(self, 
        max_grad_norm=None,
        warmup_steps=None,
        use_amp=None, 
        scaler=None,
        train_loss_dict=None):
        '''
        Default training one iteration steps, can be subclass can reimplemented.
        '''        
        skip_scheduler = False
        num_train_objectives = len(self.train_dataloader)
        for train_idx in range(num_train_objectives):
            data_iterator = self.data_iterators[train_idx]
            loss_model = self.loss_models[train_idx]
            loss_model.zero_grad()
            loss_model.train()
            optimizer = self.optimizers[train_idx]
            scheduler = self.schedulers[train_idx]

            # get a batch of data from the target data_iterator
            data = self._get_a_train_batch(data_iterator=data_iterator, train_idx=train_idx)
            data = self.prepare_input(data)

            # update model by backpropagation
            if use_amp:
                loss_value, skip_scheduler, scale_before_step = self._update_one_iteration_amp(loss_model=loss_model, data=data, optimizer=optimizer, scaler=scaler, max_grad_norm=max_grad_norm)
            else:
                loss_value = self._update_one_iteration(loss_model=loss_model, data=data, optimizer=optimizer, max_grad_norm=max_grad_norm)                

            train_loss_dict[train_idx].append(loss_value.item())

            if use_amp:
                skip_scheduler = scaler.get_scale() != scale_before_step

            if not skip_scheduler and warmup_steps > 0:
                scheduler.step()

    def _update_one_iteration(self, loss_model, data, optimizer, max_grad_norm):
        loss_model_return = loss_model(data)
        loss_value = loss_model_return['loss_value']
        loss_value.backward()

        print('loss_value', loss_value.item())

        # check if nan in the gradients of all modules in the loss_model
        # for debugging
        # for i, (name, param) in enumerate(loss_model.named_parameters()):
        #     if param.grad is not None:
        #         if torch.isnan(param.grad).any():
        #             pdb.set_trace()
        #             raise ValueError(f'Gradient of {name} is nan.')

        torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
        optimizer.step()
        return loss_value

    def _update_one_iteration_amp(self, loss_model, data, optimizer, scaler, max_grad_norm):
        with autocast():
            loss_return = loss_model(data)
        loss_value = loss_return['loss_value']
        scale_before_step = scaler.get_scale()
        scaler.scale(loss_value).backward()
        scaler.unscale_(optimizer)
        torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
        scaler.step(optimizer)
        scaler.update()
        skip_scheduler = scaler.get_scale() != scale_before_step
        return loss_value, skip_scheduler, scale_before_step



class SPOTModel(nn.Module):
    '''
    SPOT model that encodes the input data and predicts the trial outcome.
    
    Parameters
    ----------
    checkpoint: str
        The path to the checkpoint of the pretrained SPOT model.
    '''
    def __init__(self, 
        num_topics,
        n_trial_projector=2,
        n_timestemp_projector=2,
        n_rnn_layer=1,
        n_trial_per_batch=None,
        checkpoint='./pretrained_model', 
        device='cuda:0',
        **kwargs) -> None:
        super().__init__()
        self.num_topics = num_topics
        self.n_trial_per_batch = n_trial_per_batch

        self.molecule_encoder = MoleculeEncoder(input_dir=os.path.join(checkpoint, 'molecule_encoder'),
                                                    device=device)
        self.disease_encoder = DiseaseEncoder(device=device)
        self.protocol_encoder = ProtocolEncoder()
        self.feature_encoder = FeatureEncoder(1, 50)
        self.topic_encoder = TopicEncoder(num_topics, 50)

        self.rnn = nn.RNN(50, 50, num_layers=n_rnn_layer, batch_first=True, bidirectional=False)

        self.spatial_interaction = nn.MultiheadAttention(50, 1, dropout=0, bias=False, batch_first=True)
        
        self.trial_projector = Highway(50, n_trial_projector)
        self.timestep_projector = Highway(50, n_timestemp_projector)

        self.predictor = nn.Sequential(nn.ReLU(), nn.Linear(100, 50), nn.Linear(50, 1))
        self.activation = nn.ReLU()

        self.loss_fn = nn.BCEWithLogitsLoss(reduction='none')


        self.to(device)

    def forward(self, inputs, return_loss=True):
        '''
        inputs: dict
            {
                'topic': List[List[torch.Tensor]], # [batch_size, n_timestep, n_trials, emd_dim]
                'label':  List[torch.Tensor], # [batch_size, n_timestep, n_trials]
                'mask': List[torch.Tensor], # [batch_size, n_timestep, n_trials]
            }
        '''
        outputs = defaultdict(list)
        outputs['loss_value'] = 0
        for i in range(len(inputs['topic'])):
            topic_embs = pad_sequence(inputs['topic'][i], batch_first=True, padding_value=0)
            mask = inputs['mask'][i]
            mask = mask.to(topic_embs.device)
            
            # propogate historical information
            hidden_embs = topic_embs.sum(1) / mask.sum(1)[:,None] # [num_timestep, hidden_dim]
            hidden_embs = hidden_embs[None]
            hidden_embs = self.rnn(hidden_embs)[0]  # [1, num_timestep, hidden_dim]
            hidden_embs = hidden_embs.permute(1,0,2) # [num_timestep, 1, hidden_dim]

            # interaction across trials at the same timestep with the temporal embedding
            interact_embs = torch.cat([hidden_embs, topic_embs], dim=1)
            iteract_mask = torch.cat([torch.ones(mask.shape[0],1).to(mask.device), mask], 1)
            sp_hidden_embs = self.spatial_interaction(interact_embs, interact_embs, interact_embs, key_padding_mask=iteract_mask.float())[0] # [num_timestep, max_num_trials+1, hidden_dim]
            hidden_embs = sp_hidden_embs[:,1:,...] # remove the hidden state of the cls token (the first token is the temporal embedding)

            # concatenate
            hidden_embs = torch.cat([hidden_embs, topic_embs], dim=-1) # [num_timestep, max_num_trials, hidden_dim*2]

            # make predictions
            pred_logits = self.predictor(hidden_embs) # [num_timestep, max_num_trials, 1]
            outputs['pred_logits'].append(pred_logits[mask==1]) # get corresponding logits
            if 'label' in inputs and return_loss:
                label = inputs['label'][i]
                label = label.to(pred_logits.device)
                mask = mask.to(pred_logits.device)
                loss = self.compute_loss(pred_logits, label, mask)
                outputs['loss'].append(loss)
                outputs['label'].append(label[mask==1])

        if return_loss:
            all_losses = outputs.pop('loss')
            all_losses = torch.cat(all_losses, dim=0)
            if self.training and self.n_trial_per_batch is not None:
                # sampling trials for training
                n_trial = min(all_losses.shape[0], self.n_trial_per_batch)
                idx = torch.randperm(all_losses.shape[0])[:n_trial]
                outputs['loss_value'] = all_losses[idx].mean()
            else:
                outputs['loss_value'] = all_losses.mean()

        return outputs

    def prepare_input(self, inputs, return_label=True):
        '''
        transform the input data into the format that can be fed into the SPOT model
        generate the embeddings of molecule, disease, protocol serving for the sequential predictive model
        return labels for each topic
        return nctids for each topic

        Parameters
        ----------
        inputs: dict
            {
                'topic': List[List[Topic]],
                'mask': List[List[torch.Tensor]],
            }
        '''
        outputs = defaultdict(list)
        for i in range(len(inputs['topic'])):
            topic = inputs['topic'][i]
            topic_emb = self.embed_topic(topic) # [num_timestep, num_trials, emb_dim]
            outputs['topic'].append(topic_emb)
            outputs['topic_id'].append(topic.topic_id)
            if return_label:
                label = [torch.tensor([t_.attr_dict['label'] for t_ in t]) for t in inputs['topic'][i].trials]
                label = pad_sequence(label, batch_first=True, padding_value=0)
                outputs['label'].append(label)
        outputs['mask'] = inputs['mask']
        return outputs
    
    def embed_topic(self, topic):
        # embed the topic into the topic embedding
        # topic: [num_timestep, num_trials, emb_dim]
        topic_embs = []
        for trials in topic.trials: 
            ts_embs = []
            for trial in trials:
                # embed the molecule, disease, protocol
                trial_emb = self.embed_trial(trial) # [1, 50]
                # check if isnan in the embedding
                if torch.isnan(trial_emb).any():
                    pdb.set_trace()

                topic_emb = self.topic_encoder(topic.topic_id) # [1, 50]
                ts_embs.append(trial_emb+topic_emb)
            ts_embs = torch.cat(ts_embs, dim=0) # [num_trials, 50]
            ts_embs = self.timestep_projector(ts_embs) # [num_trials, 50]
            topic_embs.append(ts_embs)

        return topic_embs # [num_timestep, num_trials, 50]
    
    def embed_trial(self, trial):
        # embed the trial into the trial embedding
        # trial: Trial object
        mol_emb = self.molecule_encoder(trial.attr_dict['smiless']) # [1, 50] molecule + [1, 50] pharamakinetics
        # check if isnan in the embedding
        if torch.isnan(mol_emb).any():
            pdb.set_trace()        
        protocol_emb = self.protocol_encoder(trial.attr_dict['nctid']) # [2, 50]
        if torch.isnan(protocol_emb).any():
            pdb.set_trace()
        disease_emb = self.disease_encoder(trial.attr_dict['icdcodes']) # [1, 50]
        if torch.isnan(disease_emb).any():
            pdb.set_trace()
        year_emb = self.feature_encoder(trial.attr_dict['year']) # [1, 50]

        trial_emb = torch.cat([mol_emb, protocol_emb, disease_emb, year_emb], dim=0) # [4, 50]

        # add CLS token embedding
        trial_emb = torch.unsqueeze(trial_emb, dim=0) # [1, 4, 50]
        # trial_emb = self.cls_encoder(trial_emb)['embedding'] # [1, 5, 50]

        # highway network
        trial_emb = self.trial_projector(trial_emb) # [1, 5, 50]
        # return trial_emb[:,0,:] # get CLS embedding [1, 50]
        return trial_emb.mean(dim=1) # get mean embedding [1, 50]
    
    def compute_loss(self, pred_logits, label, mask):
        # compute the loss
        # pred_logits: [num_timestep, max_num_trials, 1]
        # label: [num_timestep, max_num_trials]
        # mask: [num_timestep, max_num_trials]
        loss = self.loss_fn(pred_logits, label.float().unsqueeze(-1))
        loss = loss * mask.unsqueeze(-1)

        # flatten the tensor of the loss
        return loss[loss!=0]

    def feed_lst_of_module(self, input_feature, lst_of_module):
        x = input_feature
        for single_module in lst_of_module:
            x = self.activation(single_module(x))
        return x

    def init_criteria_embedding(
        self, 
        input_data,
        criteria_column="criteria"
        ):
        # check if the criteria column in the input data
        assert criteria_column in input_data.df.columns, f"{criteria_column} is not in the input data."

        # initialize the criteria embedding
        _ = input_data.get_ec_sentence_embedding(criteria_column)
        # build nctid2criteria_emb assigning to self.protocol_encoder
        nct2emb = input_data.get_nct_to_ec_emb()
        self.protocol_encoder.update_criteria_emb(
            inc_emb_dict=nct2emb['nct2incemb'],
            exc_emb_dict=nct2emb['nct2excemb'],
            )

    def _generate_square_subsequent_mask(self, sz):
        mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
        mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
        return mask


class SPOT(TrialOutcomeBase):
    """
    Implement sequential predictive modeling of clinical trial outcome with meta-learning (SPOT) [1]_.

    Parameters
    ----------
    num_topics: int
        Number of topics to discover.
    
    n_trial_projector: int
        Number of layers in the trial projector.
    
    n_timestemp_projector: int
        Number of layers in the timestamp projector.
    
    n_rnn_layer: int
        Number of layers in the RNN for encoding ontology & smiles strings.
        
    criteria_column: str
        The column name of the criteria in the input data.

    batch_size: int
        Batch size for training.
    
    n_trial_per_batch: int
        Number of trials in each batch. If None, use all trials in the topic in each batch.
    
    learning_rate: float
        Learning rate for training.
    
    weight_decay: float
        Regularization strength for l2 norm; must be a positive float.
    
    epochs: int
        Number of training epochs.
    
    evaluation_steps: int
        Number of steps to evaluate the model.
    
    warmup_ratio: float
        Warmup ratio for learning rate scheduler.
    
    device: str
        Device to use for training and inference.

    seed: int
        Random seed.

    Notes
    -----
    .. [1] Wang, Z., Xiao, C., & Sun, J. (2023). SPOT: Sequential Predictive Modeling of Clinical Trial Outcome with Meta-Learning. arXiv preprint arXiv:2304.05352.
    """
    def __init__(self,
        num_topics=50,
        n_trial_projector=2,
        n_timestemp_projector=2,
        n_rnn_layer=1,
        criteria_column='criteria',
        batch_size=1,
        n_trial_per_batch=None,
        learning_rate=1e-4,
        weight_decay=1e-4,
        epochs=10,
        evaluation_steps=50,
        warmup_ratio=0,
        device="cuda:0",
        seed=42,
        output_dir="./checkpoints/spot",
        ):
        self.config = {
            'num_topics': num_topics,
            'n_trial_projector': n_trial_projector,
            'n_timestemp_projector': n_timestemp_projector,
            'n_rnn_layer': n_rnn_layer,
            'criteria_column': criteria_column,
            'batch_size': batch_size,
            'n_trial_per_batch':n_trial_per_batch,
            'learning_rate': learning_rate,
            'epochs': epochs,
            'weight_decay': weight_decay,
            'evaluation_steps':evaluation_steps,
            'warmup_ratio':warmup_ratio,
            'seed': seed,
            'output_dir': output_dir,
            'device': device,
            }
        self.model = SPOTModel(**self.config)
        self.topic_discover = TrialTopicDiscovery(num_topics=num_topics, random_seed=seed)
        self.training = False
        print(self.config)

    def fit_topic(self, df_train):
        '''
        Discover topics from the training data.

        Parameters
        ----------
        df_train: pandas.DataFrame
            A dataframe containing the training data.
        '''
        self.topic_discover.fit(df_train)

    def transform_topic(self, df, df_val=None, df_test=None):
        '''
        Transform the data into a SequenceTrial object.
        
        Parameters
        ----------
        df: pandas.DataFrame
            A dataframe containing the training data.
        
        df_val: pandas.DataFrame
            A dataframe containing the validation data.
        
        df_test: pandas.DataFrame
            A dataframe containing the test data.
        '''

        seqtrain = self.topic_discover.transform(df)
        seqval = None
        seqtest = None
        if df_val is not None:
            seqval = self.topic_discover.add_trials(seqtrain, df_val)
            if df_test is not None:
                seqtest = self.topic_discover.add_trials(seqval, df_test)        
        return {'seqtrain':seqtrain, 'seqval':seqval, 'seqtest':seqtest}

    def add_trials(self, seqtrain, df):
        '''Add new trials to the SequenceTrial object. It is used for the validation and test data.

        Parameters
        ----------
        seqtrain: model_utils.data_structure.SequenceTrial
            The sequence of clinical trials represented by SequenceTrial object.

        df: pd.DataFrame
            The dataframe of new trials that should be added to the input seqtrain.
        '''
        return self.topic_discover.add_trials(seqtrain, df)

    def fit(self, train_data, valid_data=None):
        '''Train the SPOT model. It has two/three steps:
        1. Encode the criteria using pretrained BERT to get the criteria embedding (optional).
        2. Discover topics from the training data and transform the data into a SequenceTrial object.
        3. Train the SPOT model.

        Parameters
        ----------
        train_data: TrialOutcomeDataset
            Training data, should be a `TrialOutcomeDataset` object.

        valid_data: TrialOutcomeDataset
            Validation data, should be a `TrialOutcomeDataset` object.
        '''
        # initialize the criteria embedding using pretrained language models
        self.model.init_criteria_embedding(train_data, self.config['criteria_column'])
        if valid_data is not None:
            # will add the criteria embedding for the validation data
            self.model.init_criteria_embedding(valid_data, self.config['criteria_column'])

        # initialize the topics using the topic discovery model
        self.fit_topic(train_data.df)

        seqdatas = self.transform_topic(train_data.df, valid_data.df if valid_data is not None else None)
        train_seq_data = seqdatas['seqtrain']
        valid_seq_data = seqdatas['seqval'] if valid_data is not None else None
        self._fit_model(train_seq_data, valid_seq_data)

        # save the seqdata
        self.seqdata = train_seq_data if valid_seq_data is None else valid_seq_data

    def predict(self, data, target_trials=None):
        '''
        Predict the outcome of a clinical trial.

        Parameters
        ----------
        data: TrialOutcomeDataset
            A `TrialOutcomeDataset` object containing the data to predict.
        
        target_trials: list
            A list of trial ids to predict. If None, all trials in `data` will be predicted.
        '''
        self.eval()
        if target_trials is not None:
            target_trials = data.df['nctid'].tolist()

        # need to initialize the criteria embedding for the test data
        self.model.init_criteria_embedding(data, self.config['criteria_column'])

        # add test data to the seqtrain stored in the model
        seqtest = self.add_trials(self.seqdata, data.df)

        loader = self.get_test_dataloader(seqtest)
        result = self._predict_on_dataloader(loader)
        pred, label, nctid = result['pred'], result['label'], result['nctid']
        if target_trials is not None:
            bool_idx = np.isin(nctid, target_trials)
            nctid = np.array(nctid)
            result = {'pred': pred[bool_idx], 'label': label[bool_idx], 'nctid': nctid[bool_idx]}
        
        return result

    def save_model(self, output_dir=None):
        '''
        Save the model to a directory.

        Parameters
        ----------
        output_dir: str or None
            The directory to save the model.
            If None, use the default directory './checkpoints'.
        '''
        if output_dir is None:
            output_dir = './checkpoints'
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        # save the model object into pickle
        with open(os.path.join(output_dir, constants.MODEL_OBJECT_NAME), 'wb') as f:
            pickle.dump(self.model, f)

        # save config
        with open(os.path.join(output_dir, constants.CONFIG_NAME), 'w') as f:
            f.write(json.dumps(self.config))

        # save topic discover
        with open(os.path.join(output_dir, constants.TOPIC_DISCOVER_NAME), 'wb') as f:
            pickle.dump(self.topic_discover, f)

        # save the seqdata object
        with open(os.path.join(output_dir, constants.SEQDATA_NAME), 'wb') as f:
            pickle.dump(self.seqdata, f)

        print('save model to {}'.format(output_dir))

    def load_model(self, input_dir=None):
        '''
        Load the model from a directory.

        Parameters
        ----------
        input_dir: str or None
            The directory to load the model.
        
        '''
        if input_dir is None:
            raise ValueError('Please specify the input directory.')

        self._load_config(input_dir)

        # load model object
        with open(os.path.join(input_dir, constants.MODEL_OBJECT_NAME), 'rb') as f:
            self.model = pickle.load(f)

        # load topic discover
        with open(os.path.join(input_dir, constants.TOPIC_DISCOVER_NAME), 'rb') as f:
            self.topic_discover = pickle.load(f)
        
        print('load model from {}'.format(input_dir))

    def get_train_dataloader(self, train_data):
        '''
        Build a dataloader for training a spot model.
        Parameters
        ----------
        train_data: SequenceTrial
            A dataset consisting of multiple topics of clinical trials. Each topic is a list of clinical trials under this topic ordered with their timestamps.
        '''
        loader = DataLoader(train_data, batch_size=self.config['batch_size'], shuffle=True, collate_fn=SPOTCollator())
        return loader
    
    def get_test_dataloader(self, test_data):
        '''
        Build a dataloader for testing a spot model.
        Parameters
        ----------
        test_data: SequenceTrial
            A dataset consisting of multiple topics of clinical trials. Each topic is a list of clinical trials under this topic ordered with their timestamps.
        '''
        loader = DataLoader(test_data, batch_size=1, shuffle=False, collate_fn=SPOTCollator())
        return loader
    
    def train(self, mode=True):
        self.training = mode
        self.model.train(mode)
        return self
    
    def eval(self, mode=False):
        self.training = mode
        self.model.eval()
        return self
    
    def prepare_input(self, data, return_label=False):
        return self.model.prepare_input(data, return_label=return_label)
    
    def _fit_model(self, train_data, valid_data):
        trainloader = self.get_train_dataloader(train_data)
        train_objectives = [(trainloader, self.model)]
        trainer = SPOTTrainer(
            model = self,
            train_objectives = train_objectives,
            test_data=valid_data,
            test_metric='pr_auc',
            output_dir=self.config['output_dir'],
            )
        trainer.train(**self.config)
    
    @torch.no_grad()
    def _predict_on_dataloader(self, test_dataloader):
        self.model.eval()
        pred_list, label_list, nctid_list = [], [], []
        for batch in test_dataloader:
            nctids = [collect_trials_from_topic(x) for x in batch['topic']]
            nctid_list.extend(nctids)
            inputs = self.prepare_input(batch, return_label=True)
            outputs = self.model(inputs)
            pred_list.append(torch.cat(outputs['pred_logits'], dim=0).cpu())
            label_list.append(torch.cat(outputs['label'], dim=0).cpu())

        pred = torch.cat(pred_list, dim=0)
        pred = torch.sigmoid(pred).numpy()
        label = torch.cat(label_list, dim=0).numpy()
        nctid_list = sum(nctid_list, [])
        return {'pred':pred, 'label':label, 'nctid':nctid_list}
    
    def _load_config(self, input_dir):
        with open(os.path.join(input_dir, constants.CONFIG_NAME), 'r') as f:
            config = json.loads(f.read())
        for k in constants.TRAINING_PARAMETERS:
            if k in config: config.pop(k)
        self.config.update(config)





'''
TODO: add .from_pretrained to load pretrained trial search model
from cloud storage.
'''
import abc
import os
import json
import pdb

import torch
import pandas as pd

from pytrial.utils.check import check_model_dir
from pytrial.utils.check import make_dir_if_not_exist

class TrialSearchBase(abc.ABC):
    '''Abstract class for all trial search algroithms.

    Parameters
    ----------
    experiment_id: str, optional (default = 'test')
        The name of current experiment.

    '''
    @abc.abstractmethod
    def __init__(self, experiment_id='test'):
        check_model_dir(experiment_id)
        self.checkout_dir = os.path.join('./experiments_records', experiment_id,
                                         'checkpoints')
        self.result_dir = os.path.join('./experiments_records', experiment_id,
                                       'results')
        make_dir_if_not_exist(self.checkout_dir)
        make_dir_if_not_exist(self.result_dir)


    @abc.abstractmethod
    def fit(self, train_data, valid_data):
        '''
        Fit the model with training data. Need to implement in subclass.

        Parameters
        ----------
        train_data: dict
            Training data for model fitting.

            train_data = {

            'x': pd.DataFrame,
            
            'fields': list[str],
            
            'y': pd.Series or np.array,
            
            }

        valid_data: dict
            Validation data.
            
            valid_data = {
                
            'x': pd.DataFrame,
            
            'fields': list[str],
            
            'y': pd.Series or np.array,
            
            }

        Returns
        -------
        self: object
            The trained model.

        '''
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, test_data):
        raise NotImplementedError

    @abc.abstractmethod
    def load_model(self, checkpoint):
        '''
        Parameters
        ----------
        checkpoint: str
            The path to the saved model.

        Returns
        -------
        self: object
            The loaded pretrained model.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def save_model(self, output_dir):
        '''
        Parameters
        ----------
        output_dir: str
            The directory to save the model states.

        '''
        raise NotImplementedError

    @abc.abstractmethod
    def encode(self, inputs):
        '''
        Encode input documents into embeddings.

        Parameters
        ----------
        inputs: dict
            The input documents.
        '''
        raise NotImplementedError

    def train(self, mode=True):
        self.training = mode
        self.model.train()
        return self
    
    def eval(self, mode=False):
        self.training = mode
        self.model.eval()
        return self

    @abc.abstractmethod
    def _build_model(self):
        raise NotImplementedError

    @abc.abstractmethod
    def __getitem__(self, tag):
        '''
        Get the embeddings of documents by the trial tags.

        Parameters
        ----------
        tag: str, int, list[str], list[int]
            The tag (or tags) to be looked up in the model.

        Returns
        -------
        The embeddings of each document.
        '''
        raise NotImplementedError



    def _save_checkpoint(self, state,
                        epoch_id=0,
                        is_best=False,
                        output_dir=None,
                        filename='checkpoint.pth.tar'):
        if output_dir is None:
            output_dir = self.checkout_dir

        if epoch_id < 1:
            filepath = os.path.join(output_dir, 'latest.' + filename)
        elif is_best:
            filepath = os.path.join(output_dir, 'best.' + filename)
        else:
            filepath = os.path.join(self.checkout_dir,
                                    str(epoch_id) + '.' + filename)
        torch.save(state, filepath)

    def _save_model_config(self, model_config, output_dir=None):
        if output_dir is None:
            output_dir = self.checkout_dir
        temp_path = os.path.join(output_dir, "model_config.json")
        if os.path.exists(temp_path):
            os.remove(temp_path)
        with open(temp_path, "w", encoding='utf-8') as f:
            f.write(json.dumps(model_config, indent=4))

    def _load_model_config(self, checkpoint=''):
        if checkpoint == '':
            temp_path = os.path.join(self.checkout_dir,
                                     'model_config.json')
            assert os.path.exists(
                temp_path), 'cannot find predictor_config.json, please it in dir {0}'.format(
                self.checkout_dir)
        else:
            temp_path = checkpoint
            assert os.path.exists(
                temp_path), 'cannot find checkpoint file from path: {0}'.format(
                checkpoint)
        print('load predictor config file from {0}'.format(temp_path))
        with open(temp_path, 'r') as f:
            predictor_config = json.load(f)
        return predictor_config

    def _input_data_check(self, inputs):
        '''
        Check the training / testing data fits the formats.
        Target to (1) check if inputs valid,
                    if not, give tips about the data problem.

        Parameters
        ----------
        inputs: {
                'x': pd.DataFrame,
                'fields': list[str],
                'tag': str,
                }
        '''
        # check overall input format
        assert 'x' in inputs, 'No input trial doc dataframe found in inputs.'
        df = inputs['x']
        if 'fields' in inputs:
            try:
                _ = df[inputs['fields']]
            except:
                raise Exception('Cannot find the specified `fields` in inputs dataframe.')
        if 'tag' in inputs:
            try:
                _ = df[inputs['tag']]
            except:
                raise Exception('Cannot find the specified `tag` in inputs dataframe.')

        # check data type
        try:
            _ = df.applymap(str)
        except:
            raise Exception('Cannot transform the input dataframe to str type, please check the inputs.')

    def _process_dataframe(self, df, fields):
        if fields is not None:
            df = df[fields]
        if 'nct_id' in df:
            df = df.drop(['nct_id'], axis=1)
        df = df.applymap(str)
        df = df.apply(lambda x: x.name + ': ' + x)
        df = df.applymap(lambda x: x.lower())
        df_raw_texts = df.agg(' '.join, axis=1)
        df_raw_texts = pd.DataFrame(df_raw_texts, columns=['text'])
        return df_raw_texts


def whitening_torch_final(embeddings):
    '''
    Whitening the embeddings.

    Parameters
    ----------
    embeddings: torch.Tensor
        The embeddings to be whitened. The shape is (n, d).
    '''
    mu = torch.mean(embeddings, dim=0, keepdim=True)
    cov = torch.mm((embeddings - mu).t(), embeddings - mu)
    u, s, vt = torch.svd(cov)
    W = torch.mm(u, torch.diag(1/torch.sqrt(s)))
    embeddings = torch.mm(embeddings - mu, W)
    return embeddings


import pdb
import math

import numpy as np

from pytrial.utils.trainer import Trainer

class PatientTrialTrainer(Trainer):
    def prepare_input(self, data):
        return self.model._prepare_input(data)

    def train_one_iteration(self,
        max_grad_norm=None, 
        warmup_steps=None, 
        use_amp=None,
        scaler=None,
        train_loss_dict=None):
        skip_scheduler = False
        num_train_objectives = len(self.train_dataloader)
        for train_idx in range(num_train_objectives):
            # note that for each train objective there are two dataloaders:
            # patient dataloader and trial dataloader

            patient_data_iterator = self.data_iterators[train_idx][0]
            trial_data_iterator = self.data_iterators[train_idx][1]
            loss_model = self.loss_models[train_idx]
            loss_model.zero_grad()
            loss_model.train()
            optimizer = self.optimizers[train_idx]
            scheduler = self.schedulers[train_idx]

            # prepare input for training
            patient_batch = self._get_a_train_batch(patient_data_iterator, train_idx, 0)
            trial_batch = self._get_a_train_batch(trial_data_iterator, train_idx, 1)
            patient_batch.update(trial_batch)
            inputs = self.prepare_input(patient_batch)

            # update model using input batch of data
            if use_amp:
                loss_value, skip_scheduler, scale_before_step = self._update_one_iteration_amp(loss_model=loss_model, data=inputs, optimizer=optimizer, scaler=scaler, max_grad_norm=max_grad_norm)
            else:
                loss_value = self._update_one_iteration(loss_model=loss_model, data=inputs, optimizer=optimizer, max_grad_norm=max_grad_norm)

            train_loss_dict[train_idx].append(loss_value.item())

            if use_amp:
                skip_scheduler = scaler.get_scale() != scale_before_step

            if not skip_scheduler and warmup_steps > 0:
                scheduler.step()

    def evaluate(self):
        patient_loader, trial_loader = self.test_dataloader
        res = self.model._predict_on_dataloader(patient_loader, trial_loader, return_label=True)

        # trial-level acc
        pred_trial = res['pred_trial']
        label_trial = res['label_trial']
        pred_trial = np.concatenate([p[1] for p in pred_trial])
        acc = np.sum(pred_trial == label_trial) / np.prod(label_trial.shape)

        # criteria-level acc
        inc_ec_acc, exc_ec_acc = [], []
        for i, pred_inc in enumerate(res['pred_inc']):
            pred_inc = np.argmax(pred_inc.squeeze(0), -1) # all_patient, num_ec
            label_inc = res['label_trial_inc'][i]
            label_mask = np.ones(label_inc.shape)
            label_mask[label_inc == 2] = 0 # only count on label is match or unmatch
            inc_ec_acc_ = np.sum((pred_inc == label_inc).astype(int) * label_mask) / np.sum(label_mask)
            inc_ec_acc.append(inc_ec_acc_)

        for i, pred_exc in enumerate(res['pred_exc']):
            pred_exc = np.argmax(pred_exc.squeeze(0), -1) # all_patient, num_ec
            label_exc = res['label_trial_exc'][i]
            label_mask = np.ones(label_exc.shape)
            label_mask[label_exc == 2] = 0 # only count on label is match or unmatch
            exc_ec_acc_ = np.sum((pred_exc == label_exc).astype(int) * label_mask) / np.sum(label_mask)
            exc_ec_acc.append(exc_ec_acc_)

        return {
            'trial-level-acc':acc, 
            'inc-level-acc':np.mean(inc_ec_acc), 
            'exc-level-acc':np.mean(exc_ec_acc)
            }

    def get_test_dataloader(self, test_data):
        return self.model.get_test_dataloader(test_data)
    
    def _build_data_iterator(self, epochs):
        self.data_iterators = [[iter(dataloader_) for dataloader_ in dataloader] for dataloader in self.train_dataloader]
        steps_per_epoch = max([len(dataloader_) for dataloader_ in self.train_dataloader[0]])
        num_train_steps = int(steps_per_epoch * epochs)
        return  steps_per_epoch, num_train_steps

    def _get_a_train_batch(self, data_iterator, train_idx, data_idx):
        try:
            data = next(data_iterator)

        except StopIteration:
            data_iterator = iter(self.train_dataloader[train_idx][data_idx])
            self.data_iterators[train_idx][data_idx] = data_iterator
            data = next(data_iterator)
            
        return data


import torch
from torch import nn
from torch.utils.data import DataLoader
from torch.nn.utils.rnn import pad_sequence, pack_padded_sequence, pad_packed_sequence

import numpy as np
from collections import defaultdict

from pytrial.data.patient_data import SequencePatientBase, SeqPatientCollator
from pytrial.utils.check import (
    check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
)
from .base import SequenceSimulationBase
from ..losses import MultilabelBinaryXentLossWithKLDivergence
from ..trainer import SeqSimEvaTrainer
from ..data import SequencePatient

class CausalConv1d(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, dilation=1, **kwargs):
        super(CausalConv1d, self).__init__()
        self.pad = (kernel_size - 1) * dilation
        self.conv = nn.Conv1d(in_channels, out_channels, kernel_size, padding=self.pad, dilation=dilation, **kwargs)

    def forward(self, input):
        return self.conv(input)[:,:,:-self.conv.padding[0]]

def connector(mu, log_var):
    std = torch.exp(0.5 * log_var)
    eps = torch.randn_like(std)
    return eps * std + mu

class Encoder(nn.Module):
    def __init__(self, emb_size, latent_dim, n_rnn_layer, total_vocab_size):
        super(Encoder, self).__init__()
        self.hidden_dim = emb_size
        self.embedding_matrix = nn.Linear(total_vocab_size, emb_size, bias=False)
        self.lstm = nn.LSTM(input_size=emb_size,
                            hidden_size=emb_size,
                            num_layers=n_rnn_layer,
                            bidirectional=True,
                            batch_first=True)
        self.latent_encoder = nn.Linear(2*emb_size, 2*latent_dim)

    def forward(self, input, lengths):
        visit_emb = self.embedding_matrix(input)
        packed_input = pack_padded_sequence(visit_emb, lengths, batch_first=True, enforce_sorted=False)
        packed_output, _ = self.lstm(packed_input)
        output, _ = pad_packed_sequence(packed_output, batch_first=True)
        out_forward = output[range(len(output)), [l - 1 for l in lengths], :self.hidden_dim]
        out_reverse = output[:, 0, self.hidden_dim:]
        out_combined = torch.cat((out_forward, out_reverse), 1)
        mean_logvar = self.latent_encoder(out_combined)
        return mean_logvar

class Decoder(nn.Module):
    def __init__(self, emb_size, latent_dim, total_vocab_size, max_visit):
        super(Decoder, self).__init__()
        self.max_visit = max_visit
        self.deconv_list = nn.ModuleList([nn.ConvTranspose1d(latent_dim if i == 0 else emb_size, emb_size, 3, stride=3) for i in range(int(np.ceil(np.power(max_visit, 1/3))))])
        self.causal_conv1 = CausalConv1d(emb_size, emb_size, 5, dilation=2)
        self.causal_conv2 = CausalConv1d(emb_size, 2*emb_size, 5, dilation=2)
        self.causal_conv3 = CausalConv1d(2*emb_size, total_vocab_size, 5, dilation=2)

    def forward(self, input):
        out = input.unsqueeze(2)
        for deconv in self.deconv_list:
          out = deconv(out)
        out = out[:,:,:self.max_visit]
        out = self.causal_conv1(out)
        out = self.causal_conv2(out)
        out = self.causal_conv3(out)
        out = out.transpose(1, 2)
        return out
      
class BuildModel(nn.Module):
    def __init__(self,
        max_visit,
        emb_size,
        latent_dim,
        n_rnn_layer,
        total_vocab_size,
        device,
        **kwargs,
        ) -> None:
        super().__init__()
        
        self.latent_dim = latent_dim
        self.device = device

        self.encoder_module = Encoder(
            emb_size=emb_size, 
            latent_dim=latent_dim, 
            n_rnn_layer=n_rnn_layer, 
            total_vocab_size=total_vocab_size,
        )

        self.decoder_module = Decoder(
            emb_size=emb_size, 
            latent_dim=latent_dim, 
            total_vocab_size=total_vocab_size, 
            max_visit=max_visit
        )
      
    def forward(self, inputs):
        x, input_lengths = inputs['v'], inputs['v_lengths']
        mean_logvar = self.encoder_module(x, input_lengths)
        mu = mean_logvar[:,:self.latent_dim]
        log_var = mean_logvar[:,self.latent_dim:]
        kl_loss = torch.mean(-0.5 * torch.sum(1 + log_var - mu ** 2 - log_var.exp(), dim = 1), dim = 0)
        decoder_inputs = connector(mu, log_var)
        code_logits = self.decoder_module(decoder_inputs)
        for idx in range(len(inputs['v_lengths'])):
            code_logits[idx,inputs['v_lengths'][idx]:] = 0
        return code_logits[:,:inputs['v'].size(1)], inputs['v'], kl_loss
      
    def sample(self, n_samples):
        decoder_inputs = torch.randn((n_samples, self.latent_dim)).to(self.device)
        code_logits = self.decoder_module(decoder_inputs)
        sig = nn.Sigmoid()
        code_probs = sig(code_logits)
        patient_records = torch.bernoulli(code_probs)
        return patient_records
    
    def sample_from(self, x, input_lengths, n_per_sample):
        mean_logvar = self.encoder_module(x, input_lengths)
        mu = mean_logvar[:,:self.latent_dim]
        log_var = mean_logvar[:,self.latent_dim:]
        code_probs = []
        sig = nn.Sigmoid()
        for _ in range(n_per_sample):
            decoder_inputs = connector(mu, log_var)
            code_logits = self.decoder_module(decoder_inputs)
            code_probs.append(sig(code_logits))
        code_probs = torch.cat(code_probs)
        patient_records = torch.bernoulli(code_probs)
        return patient_records

class EVA(SequenceSimulationBase):
    '''
    Implement a VAE based model for longitudinal patient records simulation [1]_.
    
    Parameters
    ----------
    vocab_size: list[int]
        A list of vocabulary size for different types of events, e.g., for diagnosis, procedure, medication.

    order: list[str]
        The order of event types in each visits, e.g., ``['diag', 'prod', 'med']``.
        Visit = [diag_events, prod_events, med_events], each event is a list of codes.

    max_visit: int
        Maximum number of visits.

    emb_size: int
        Embedding size for encoding input event codes.
        
    latent_dim: int
        Size of final latent dimension between the encoder and decoder
        
    n_rnn_layer: int
        Number of RNN layers for encoding historical events.

    learning_rate: float
        Learning rate for optimization based on SGD. Use torch.optim.Adam by default.

    batch_size: int
        Batch size when doing SGD optimization.

    epochs: int
        Maximum number of iterations taken for the solvers to converge.

    num_worker: int
        Number of workers used to do dataloading during training.

    Notes
    -----
    .. [1] Biswal, S., et al. (2020, December). EVA: Generating Longitudinal Electronic Health Records Using Conditional Variational Autoencoders.
    '''
    def __init__(self,
        vocab_size,
        order,
        max_visit=20,
        emb_size=64,
        latent_dim=32,
        n_rnn_layer=2,
        learning_rate=1e-3,
        batch_size=64,
        epochs=20,
        num_worker=0,
        device='cpu',# 'cuda:0',
        experiment_id='trial_simulation.sequence.eva',
        ):
        super().__init__(experiment_id)
        self.config = {
            'vocab_size':vocab_size,
            'max_visit':max_visit,
            'emb_size':emb_size,
            'latent_dim':latent_dim,
            'n_rnn_layer':n_rnn_layer,
            'device':device,
            'learning_rate':learning_rate,
            'batch_size':batch_size,
            'epochs':epochs,
            'num_worker':num_worker,
            'orders':order,
            }
        self.config['total_vocab_size'] = sum(vocab_size)
        self.device = device
        self._build_model()
        
    def _build_model(self):
        self.model = BuildModel(
            max_visit=self.config['max_visit'],
            emb_size=self.config['emb_size'],
            latent_dim=self.config['latent_dim'],
            n_rnn_layer=self.config['n_rnn_layer'],
            total_vocab_size=self.config['total_vocab_size'],
            orders=self.config['orders'],
            device=self.device
            )
        self.model = self.model.to(self.device)
        
    def fit(self, train_data):
        '''
        Train model with sequential patient records.

        Parameters
        ----------
        train_data: SequencePatientBase
            A `SequencePatientBase` contains patient records where 'v' corresponds to 
            visit sequence of different events.
        '''
        self._input_data_check(train_data)
        self._fit_model(train_data)
    
    def predict(self, n, return_tensor=False):
        '''
        Generate synthetic records

        Parameters
        ----------
        n: int
            How many samples in total will be generated.
        
        return_tensor: bool
            - If `True`, return output generated records in tensor format (n, n_visit, n_event), good for later predictive modeling.
            - If `False, return records in `SequencePatient` format.
        '''
        assert isinstance(n, int), 'Input `n` should be integer.'
        outputs = self._predict(n)
        if not return_tensor:
            outputs = self._translate_sparse_visits_to_dense(outputs)
        return outputs

    def save_model(self, output_dir):
        '''
        Save the learned simulation model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir
            
        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({
                  'encoder': self.model.encoder_module.state_dict(),
                  'decoder': self.model.decoder_module.state_dict()
              }, output_dir=output_dir)

    def load_model(self, checkpoint):
        '''
        Load model and the pre-encoded trial embeddings from the given
        checkpoint dir.
        
        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.

            - If a directory, the only checkpoint file `*.pth.tar` will be loaded.
            - If a filepath, will load from this file.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename, map_location=self.config['device'])
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        self.model.encoder_module.load_state_dict(state_dict['encoder'])
        self.model.decoder_module.load_state_dict(state_dict['decoder'])
    
    def get_train_dataloader(self, train_data):
        dataloader = DataLoader(train_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=True,
            shuffle=True,
            collate_fn=SeqPatientCollator(
                config={
                    'visit_mode':train_data.metadata['visit']['mode'],
                    'label_mode':train_data.metadata['label']['mode'],
                    }
                ),
            )
        return dataloader
    
    def get_test_dataloader(self, train_data):
        dataloader = DataLoader(train_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=False,
            shuffle=False,
            collate_fn=SeqPatientCollator(
                config={
                    'visit_mode':train_data.metadata['visit']['mode'],
                    'label_mode':train_data.metadata['label']['mode'],
                    }
                ),
            )
        return dataloader
    
    def _build_loss_model(self):
        return MultilabelBinaryXentLossWithKLDivergence(self.model)

    def _fit_model(self, train_data):
        train_dataloader = self.get_train_dataloader(train_data)
        loss_model = self._build_loss_model()
        train_objectives = [(train_dataloader, loss_model)]
        trainer = SeqSimEvaTrainer(
            model=self,
            train_objectives=train_objectives
            )
        trainer.train(**self.config)

    @torch.no_grad()
    def _predict(self, n):
        return self.model.sample(n)
    
    @torch.no_grad()
    def _predict_on_dataloader(self, test_dataloader, n, n_per_sample):
        data_iterator = iter(test_dataloader)
        outputs = []
        for data in data_iterator:
            outputs.append(self.model.sample_from(data['v'], data['v_lengths'], n_per_sample))
            
        return torch.cat(outputs)[:n]

    def _translate_sparse_visits_to_dense(self, visits):
        def _map_func(x):
            res = np.where(x > 0)[0].tolist()
            return [0] if len(res) == 0 else res # pad if nothing happened

        outputs = defaultdict(list)
        for batchv in visits:
            voc_offset = 0
            for i, o in enumerate(self.config['orders']):
                voc_size = self.config['vocab_size'][i]
                visit = batchv[...,voc_offset:voc_offset+voc_size]
                visit = visit.cpu().numpy()
                res = list(map(_map_func, visit))
                outputs[o].append(res)
                voc_offset += voc_size

        n_total = len(outputs[o])
        sample_list = []        
        for i in range(n_total):
            sample = []
            for numv in range(len(outputs[o][i])):
                visit = []
                for o in self.config['orders']:
                    visit.append(outputs[o][i][numv])
                sample.append(visit)
            sample_list.append(sample)

        # create seqpatient data
        return SequencePatient(
            data={'v':sample_list},
            metadata={
                'visit':{'mode':'dense','order':self.config['orders']},
                }
            )

    def _translate_dense_visits_to_sparse(self, visits):
        total_vocab_size = sum(self.config['vocab_size'])
        num_visits = len(visits[self.config['orders'][0]])
        outputs = np.zeros((num_visits, total_vocab_size))

        for i, o in enumerate(self.config['orders']):
            for j in range(num_visits):
                raw = visits[o][j]
                if isinstance(raw, torch.Tensor): raw = raw.detach().cpu().numpy()
                if i > 0: 
                    voc_size = sum(self.config['vocab_size'][:i-1])
                    if isinstance(raw, list):
                        raw = [r + voc_size for r in raw]
                    else:
                        raw += voc_size
                outputs[j, raw] = 1

        return outputs

    def _pad_multiple_tensor_visits(self, visits):
        new_list = []
        for v in visits:
            new_list.extend([torch.tensor(x).squeeze(0) for x in np.array_split(v, len(v))])
        return pad_sequence(new_list, batch_first=True)
    
    def _prepare_input(self, data):
        '''
        Prepare inputs for sequence simulation models.

        Parameters
        ----------
        data: dict[list]
            A batch of patient records.
        '''
        visits = data['v']
        feature = data['x']
        if not isinstance(feature, torch.Tensor): feature = torch.tensor(feature)

        feature = feature.to(self.device)

        inputs = {
            'v':{},
            'v_lengths':[],
            'x':feature, # baseline feature
            }
        
        v_lengths = [len(visits[self.config['orders'][0]][idx][:self.config['max_visit']]) for idx in range(len(visits[self.config['orders'][0]]))]
        inputs['v_lengths'] = v_lengths

        v = torch.zeros(len(v_lengths), max(v_lengths), self.config['total_vocab_size'])
        for idx in range(len(v_lengths)):
            v[idx,:v_lengths[idx]] = torch.tensor(self._translate_dense_visits_to_sparse({k: visits[k][idx][:self.config['max_visit']] for k in visits}))
        v = v.to(self.device)
        inputs['v'] = v

        return inputs
    
    def _input_data_check(self, inputs):
        assert isinstance(inputs, SequencePatientBase), f'`trial_simulation.sequence` models require input training data in `SequencePatientBase`, find {type(inputs)} instead.'
        


import joblib
import os
import json

from rdkit import Chem 
from rdkit.Chem import AllChem
from rdkit.Chem import DataStructs
import xgboost
import numpy as np 

from .base import TrialOutcomeBase 
from .model_utils.utils import trial_collate_fn

class XGBoost(TrialOutcomeBase):
    '''
    Implement XGBoost model for clinical trial outcome prediction. 

    Parameters
    ----------
    n_estimators: int
        number of trees in the forest
    
    max_depth: int
        maximum depth of the tree
    
    reg_lambda: float
        L2 regularization term on weights
    
    eval_metric: {'auc', 'logloss'}
        evaluation metric for validation data, default is AUC.
    '''
    def __init__(self,
        n_estimators=100,
        max_depth=3,
        reg_lambda=0,
        eval_metric='auc',
        ):
        super(XGBoost, self).__init__()
        self.model = xgboost.XGBClassifier(n_estimators=n_estimators, 
            max_depth=max_depth, reg_lambda=reg_lambda, eval_metric=eval_metric)
        self.config = {'n_estimators': n_estimators, 'max_depth': max_depth, 'reg_lambda': reg_lambda, 'eval_metric': eval_metric}

    def feature(self, data_loader):
        labels = []
        features = []
        nctids = []
        for nctid_lst, label_vec, smiles_lst2, icdcode_lst3, criteria_lst in data_loader:
            labels.extend(label_vec.tolist())
            nctids.extend(nctid_lst)
            for smiles_lst in smiles_lst2: 
                mol_lst = [Chem.MolFromSmiles(smiles) for smiles in smiles_lst]
                mol_lst = list(filter(lambda x:x is not None, mol_lst))
                fp_lst = [AllChem.GetMorganFingerprintAsBitVect(mol, 2, nBits=512) for mol in mol_lst]
                feature = np.zeros(512)
                for fp in fp_lst:
                    arr = np.zeros((0,), dtype=np.int8)
                    DataStructs.ConvertToNumpyArray(fp,arr)
                    feature += arr 
                feature = feature.reshape(1,-1)
                features.append(feature)
        features = np.concatenate(features, 0)
        labels = np.array(labels)	
        return nctids, features, labels 	

    def fit(self, train_data, valid_data=None):
        '''
        Train model to predict clinical trial outcomes.

        Parameters
        ----------
        train_data: TrialOutcomeDatasetBase
            Training data, should be a `TrialOutcomeDatasetBase` object.
        
        valid_data: TrialOutcomeDatasetBase
            Validation data, should be a `TrialOutcomeDatasetBase` object. 
        '''
        eval_set = None

        # build dataloader using train_data
        train_loader = self._build_dataloader_from_dataset(train_data, 
            num_workers=0, batch_size=32, shuffle=True,
            collate_fn=trial_collate_fn)
        _, features, labels = self.feature(train_loader)

        if valid_data is not None:
            valid_loader = self._build_dataloader_from_dataset(valid_data, 
                num_workers=0, batch_size=32, shuffle=False,
                collate_fn=trial_collate_fn)
            _, valid_features, valid_labels = self.feature(valid_loader)
            eval_set = [(valid_features, valid_labels)]

        self.model.fit(features, labels, eval_set=eval_set, verbose=True)

    def predict(self, test_data):
        '''
        Make clinical trial outcome predictions.
        
        Parameters
        ----------
        test_data: TrialOutcomeDatasetBase
            Testing data, should be a `TrialOutcomeDatasetBase` object.
        '''
        # build dataloader using test_data
        testloader = self._build_dataloader_from_dataset(test_data, 
            num_workers=0, batch_size=32, shuffle=False,
            collate_fn=trial_collate_fn)

        nctids, features, labels = self.feature(testloader)
        ypred = self.model.predict_proba(features)
        ypred = ypred[:, 1].tolist() 
        prediction = list(zip(nctids, ypred))
        return prediction

    def save_model(self, output_dir=None):
        '''
        Save the learned model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The output folder to save the learned model.
            If set None, will save model to `save_model/model.ckpt`.
        '''
        if output_dir is None:
            output_dir = 'save_model'
        
        if not os.path.exists(output_dir): os.makedirs(output_dir)
        
        filename = os.path.join(output_dir, 'model.ckpt')
        joblib.dump(self.model, filename)

        config_filename = os.path.join(output_dir, 'config.json')
        with open(config_filename, 'w') as f:
            json.dump(self.config, f)


    def load_model(self, checkpoint=None):
        '''
        Load the learned model from the disk.

        Parameters
        ----------
        checkpoint: str 
            The checkpoint folder to load the learned model.
            The checkpoint under this folder should be `model.ckpt`.
        '''
        if checkpoint is None:
            ckpt_dir = 'save_model'
            checkpoint = os.path.join(ckpt_dir, 'model.ckpt')
        else:
            checkpoint = os.path.join(checkpoint, 'model.ckpt')
        self.model = joblib.load(checkpoint)

        ckpt_dir = os.path.dirname(checkpoint)
        config_filename = os.path.join(ckpt_dir, 'config.json')
        with open(config_filename, 'r') as f:
            self.config = json.load(f)


import pdb
from collections import defaultdict
import joblib
import random
import os

from tqdm import tqdm
import pandas as pd
import numpy as np
from sklearn.neighbors import NearestNeighbors
from sklearn.manifold import TSNE
import numpy as np

from pytrial.data.patient_data import SequencePatientBase, SeqPatientCollator
from pytrial.utils.check import check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
from ..data import SequencePatient
from .base import SequenceSimulationBase
from .base import transform_sequence_to_table
from .base import transform_table_to_sequence


def get_neighbors(x, knn, idx, k):
    # knn = NearestNeighbors(n_neighbors=k, algorithm='auto')
    vector = x[idx]
    return knn.kneighbors([vector], k, return_distance=False)

def generate_data(df, idx):
    neighbor_ids = get_neighbors(idx)
    new_sample = [df.iloc[idx, 0], df.iloc[idx, 1], df.iloc[idx, 2]]
    for j in range(3, len(df.columns)):
        new_sample.append(df.iloc[random.choice(neighbor_ids[0]), j])
    return new_sample


class BuildModel:
    def __new__(self, config):
        model = NearestNeighbors(n_neighbors=config['k'], algorithm='auto')
        return model

class KNNSampler(SequenceSimulationBase):
    '''
    Implement an RNN based GAN model for longitudinal patient records simulation. The GAN part was proposed by Beigi et al. [1]_ to generate synthetic tabular patient records.
    We adapt it to generate sequential records by finding the nearest neighbor on the visit-level.

    Parameters
    ----------
    order: list[str]
        The order of event types in each visits, e.g., ``['diag', 'prod', 'med']``.
        Visit = [diag_events, prod_events, med_events], each event is a list of codes.

    voc: dict
        A dictionary that maps the event type to the corresponding vocabulary.

    k: int
        Number of nearest neighbors to use for synthetic data generation.
    
    tsne_dims: int
        Dimension of t-SNE embedding. Used to encode the visit events into a vector.

    Notes
    -----
    .. [1] Beigi, M., Shafquat, A., Mezey, J., & Aptekar, J. W. (2022) Synthetic Clinical Trial Data while Preserving Subject-Level Privacy. In NeurIPS 2022 Workshop on Synthetic Data for Empowering ML Research.
    '''
    def __init__(self,
        order=None,
        voc=None,
        k=3,
        tsne_dims=2,
        experiment_id='trial_simulation.sequence.knn',
        ):
        super().__init__(experiment_id)
        self.config = {
            'k': k,
            'tsne_dims': tsne_dims,
            'order':order,
            'voc': voc,
        }
        self._build_model()
        assert order is not None, 'Please specify the order of event types in each visit.'
        assert voc is not None, 'Please specify the vocabulary event types.'


    def fit(self, train_data, val_data=None):
        '''
        Fit KNN model on the input training sequential patient records.

        Parameters
        ----------
        train_data: SequencePatient
            A `SequencePatient` contains patient records where 'v' corresponds to
            visit sequence of different events.

        val_data: None
            Not used in this model. Only for compatibility with other models.
        '''
        self._input_data_check(train_data)
        self._fit_model(train_data)

    def load_model(self, checkpoint=None):
        '''
        Load the learned model from the disk.

        Parameters
        ----------
        checkpoint: str
            - If a directory, the only checkpoint file `.model` will be loaded.
            - If a filepath, will load from this file;
            - If None, will load from `self.checkout_dir`.
        '''
        if checkpoint is None:
            checkpoint = self.checkout_dir
        checkpoint_filename = check_checkpoint_file(checkpoint, suffix='model')
        model = joblib.load(checkpoint_filename)
        self.__dict__.update(model.__dict__)

    def save_model(self, output_dir=None):
        '''
        Save the fitted model and the raw data to the disk for later use.

        Parameters
        ----------
        output_dir: str
            The directory to save the model and data. If `None`, save to the default directory.
            `self.checkout_dir`.
        '''
        if output_dir is None:
            output_dir = self.checkout_dir
        make_dir_if_not_exist(output_dir)
        ckpt_path = os.path.join(output_dir, 'knnsampler.model')
        joblib.dump(self, ckpt_path)
    
    def predict(self, n_per_sample=None, return_tensor=False):
        '''
        Generate synthetic patient records by sampling and perturbing the nearest neighbors of the input patient records.

        Parameters
        ----------        
        n_per_sample: int
            Number of synthetic records to generate for each patient. If `None`, generate one synthetic record for each patient.
        
        return_tensor: bool
            If `True`, return the synthetic records as a tensor format (n, n_visit, n_event), good for later predictive modeling.
            If `False`, the synthetic records are in the same format as the input patient records.
        '''
        if n_per_sample is None:
            n_per_sample = 1
        assert isinstance(n_per_sample, int), 'Please specify the number of synthetic records to generate for each patient. Should be an integer. Get {} instead.'.format(type(n_per_sample))
        output = self._generate(n_per_sample)

        if return_tensor:
            # transform the output into tensor format
            output = self._transform_to_tensor(output)
        else:
            # transform the output into SequencePatient format
            output = self._transform_to_sequence_patient(output)
        return output

    def _build_model(self):
        self.knn = BuildModel(self.config)
        self.tsne = TSNE(n_components=self.config['tsne_dims'], random_state=0)
        self.data = None

    def _fit_model(self, train_data):
        # transform visit sequence into tabular format
        visit_df = transform_sequence_to_table(train_data.visit, 
            self.config['order'], self.config['voc'])
        visit_df.set_index(['pid', 'vid'], inplace=True)

        # fit TSNE model
        self.tsne.fit(visit_df)

        # transform visit sequence into vector
        z = self.tsne.embedding_

        # fit KNN model
        self.knn.fit(z)

        # save the fitted data for later use of generating synthetic data
        self.data = visit_df

        # save the other information of the input data
        self._input_data = train_data

    def _generate(self, n_per_sample):
        # find the nearest neighbors
        z = self.tsne.embedding_
        df = self.data.copy().reset_index()
        syn_df = pd.DataFrame(columns= df.columns)

        for j in range(n_per_sample):
            for i in tqdm(range(len(z)), desc='Generating synthetic data {}/{} per patient'.format(j+1, n_per_sample)):
                neighbor_ids = get_neighbors(z, self.knn, i, self.config['k'])
                # generate new sample
                # randomly sample one of the neighbors for each column
                new_sample = [df.iloc[i, 0], df.iloc[i, 1]]
                df_neighbors = df.iloc[neighbor_ids[0]]
                sampled = df_neighbors.apply(lambda x: np.random.choice(x), axis=0)
                new_sample.extend(sampled[2:].values)
                syn_df.loc[len(syn_df)] = new_sample

        return syn_df

    def _transform_to_tensor(self, df):
        # transform the output into tensor format
        # [n, n_visit, n_event]
        df['vid'] = df['vid'].astype(int)
        df['pid'] = df['pid'].astype(int)
        outputs = np.zeros((df['pid'].nunique(), df['vid'].max()+1, len(df.columns)-2))
        df.set_index(['pid', 'vid'], inplace=True)
        for pid in df.index.get_level_values(0).unique():
            vals = df.loc[pid].values
            outputs[pid, :vals.shape[0], :] = vals
        return outputs

    def _transform_to_sequence_patient(self, df):
        # transform the output into SequencePatient format
        df['vid'] = df['vid'].astype(int)
        df['pid'] = df['pid'].astype(int)
        visits, pids = transform_table_to_sequence(df, self.config['order'], self.config['voc'])

        # like pid to self._input_data's features
        labels = self._input_data.__dict__.get('label', None)
        if labels is not None:
            labels = [labels[pid] for pid in pids]
        features = self._input_data.__dict__.get('feature', None)
        if features is not None:
            features = [features[pid] for pid in pids]
        
        output = SequencePatient(
            data={
                'v': visits,
                'y': labels,
                'x': features,
            },
            metadata={
                'visit': {'mode':'dense'},
                'label': {'mode':'tensor'},
                'voc': self.config['voc'],
                'n_num_features': self._input_data.metadata.get('n_num_features', None),
                'cat_cardinalities': self._input_data.metadata.get('cat_cardinalities', None),
            }
        )
        return output

    def _input_data_check(self, inputs):
        assert isinstance(inputs, SequencePatientBase), f'`trial_simulation.sequence` models require input training data in `SequencePatientBase`, find {type(inputs)} instead.'



'''
input:
	data/raw_data.csv

output:
	data/sentence2embedding.pkl (preprocessing)
	protocol_embedding 
'''

import csv, pickle
import pandas as pd
from functools import reduce
from tqdm import tqdm 
import torch 
torch.manual_seed(0)
from torch import nn 
import torch.nn.functional as F

def clean_protocol(protocol):
	protocol = protocol.lower()
	protocol_split = protocol.split('\n')
	filter_out_empty_fn = lambda x: len(x.strip())>0
	strip_fn = lambda x:x.strip()
	protocol_split = list(filter(filter_out_empty_fn, protocol_split))	
	protocol_split = list(map(strip_fn, protocol_split))
	return protocol_split 

def get_all_protocols():
	input_file = 'data/raw_data.csv'
	with open(input_file, 'r') as csvfile:
		rows = list(csv.reader(csvfile, delimiter = ','))[1:]
	protocols = [row[9] for row in rows]
	return protocols

def split_protocol(protocol):
	protocol_split = clean_protocol(protocol)
	inclusion_idx, exclusion_idx = len(protocol_split), len(protocol_split)	
	for idx, sentence in enumerate(protocol_split):
		if "inclusion" in sentence:
			inclusion_idx = idx
			break
	for idx, sentence in enumerate(protocol_split):
		if "exclusion" in sentence:
			exclusion_idx = idx 
			break 		
	if inclusion_idx + 1 < exclusion_idx + 1 < len(protocol_split):
		inclusion_criteria = protocol_split[inclusion_idx:exclusion_idx]
		exclusion_criteria = protocol_split[exclusion_idx:]
		if not (len(inclusion_criteria) > 0 and len(exclusion_criteria) > 0):
			print(len(inclusion_criteria), len(exclusion_criteria), len(protocol_split))
			exit()
		return inclusion_criteria, exclusion_criteria ## list, list 
	else:
		return protocol_split, 

def collect_cleaned_sentence_set():
	protocol_lst = get_all_protocols() 
	cleaned_sentence_lst = []
	for protocol in protocol_lst:
		result = split_protocol(protocol)
		cleaned_sentence_lst.extend(result[0])
		if len(result)==2:
			cleaned_sentence_lst.extend(result[1])
	return set(cleaned_sentence_lst)

def save_trial_criteria_bert_dict_pkl():
    from pytrial.model_utils.bert import BERT
    bertmodel = BERT(device='cuda:0')
    inc_nct_2_embedding = dict()
    exc_nct_2_embedding = dict()
    input_file = constants.RAW_DATA_FILENAME
    df = pd.read_csv(input_file)
    df['criteria'] = df['criteria'].apply(str)
    ecs = df['criteria'].apply(split_protocol)
    df['inclusion_criteria'] = ecs.apply(lambda x: x[0])
    df['exclusion_criteria'] = ecs.apply(lambda x: x[1] if len(x) > 1 else ['[PAD]'])
    df.drop(columns=['criteria'], inplace=True)

    # encode inclusion
    for i, row in tqdm(df.iterrows(), total=df.shape[0]):
        inc_nct_2_embedding[row['nctid']] = bertmodel.encode(row['inclusion_criteria']).mean(axis=0).cpu().numpy()
        exc_nct_2_embedding[row['nctid']] = bertmodel.encode(row['exclusion_criteria']).mean(axis=0).cpu().numpy()
    pickle.dump(inc_nct_2_embedding, open(constants.INC_EMB_FILENAME, 'wb'))
    pickle.dump(exc_nct_2_embedding, open(constants.EXC_EMB_FILENAME, 'wb'))
    return


def save_sentence_bert_dict_pkl():
	cleaned_sentence_set = collect_cleaned_sentence_set() 
	from biobert_embedding.embedding import BiobertEmbedding
	biobert = BiobertEmbedding()
	def text2vec(text):
		return biobert.sentence_vector(text)
	protocol_sentence_2_embedding = dict()
	for sentence in tqdm(cleaned_sentence_set):
		protocol_sentence_2_embedding[sentence] = text2vec(sentence)
	pickle.dump(protocol_sentence_2_embedding, open('data/sentence2embedding.pkl', 'wb'))
	return

def load_sentence_2_vec():
	sentence_2_vec = pickle.load(open('data/sentence2embedding.pkl', 'rb'))
	return sentence_2_vec 

def protocol2feature(protocol, sentence_2_vec):
	result = split_protocol(protocol)
	inclusion_criteria, exclusion_criteria = result[0], result[-1]
	inclusion_feature = [sentence_2_vec[sentence].view(1,-1) for sentence in inclusion_criteria if sentence in sentence_2_vec]
	exclusion_feature = [sentence_2_vec[sentence].view(1,-1) for sentence in exclusion_criteria if sentence in sentence_2_vec]
	if inclusion_feature == []:
		inclusion_feature = torch.zeros(1,768)
	else:
		inclusion_feature = torch.cat(inclusion_feature, 0)
	if exclusion_feature == []:
		exclusion_feature = torch.zeros(1,768)
	else:
		exclusion_feature = torch.cat(exclusion_feature, 0)
	return inclusion_feature, exclusion_feature 


class Protocol_Embedding(nn.Sequential):
	def __init__(self, output_dim, highway_num, device ):
		super(Protocol_Embedding, self).__init__()	
		self.input_dim = 768  
		self.output_dim = output_dim 
		self.highway_num = highway_num 
		self.fc = nn.Linear(self.input_dim*2, output_dim)
		self.f = F.relu
		self.device = device 
		self = self.to(device)

	def forward_single(self, inclusion_feature, exclusion_feature):
		## inclusion_feature, exclusion_feature: xxx,768 
		inclusion_feature = inclusion_feature.to(self.device)
		exclusion_feature = exclusion_feature.to(self.device)
		inclusion_vec = torch.mean(inclusion_feature, 0)
		inclusion_vec = inclusion_vec.view(1,-1)
		exclusion_vec = torch.mean(exclusion_feature, 0)
		exclusion_vec = exclusion_vec.view(1,-1)
		return inclusion_vec, exclusion_vec 

	def forward(self, in_ex_feature):
		result = [self.forward_single(in_mat, ex_mat) for in_mat, ex_mat in in_ex_feature]
		inclusion_mat = [in_vec for in_vec, ex_vec in result]
		inclusion_mat = torch.cat(inclusion_mat, 0)  #### 32,768
		exclusion_mat = [ex_vec for in_vec, ex_vec in result]
		exclusion_mat = torch.cat(exclusion_mat, 0)  #### 32,768 
		protocol_mat = torch.cat([inclusion_mat, exclusion_mat], 1)
		output = self.f(self.fc(protocol_mat))
		return output 

	@property
	def embedding_size(self):
		return self.output_dim 



if __name__ == "__main__":
	# protocols = get_all_protocols()
	# split_protocols(protocols)
	save_sentence_bert_dict_pkl() 














import os
import json
import pdb
import wget
import dill

from .dataset import MimicDataset
from .import constants

# def load_demo_data(input_dir='./demo_data'):
#     '''
#     Load demo data.
#     '''
#     return_dict = {}

#     with open(os.path.join(input_dir, 'token_dict.json'), 'r', encoding='utf-8') as f:
#         token_dict = json.loads(f.read())
    
#     return_dict['token_dict'] = token_dict

#     with open(os.path.join(input_dir, 'cat_cardinalities.txt'), 'r', encoding='utf-8') as f:
#         cats = [l.strip() for l in f.readlines()]
#     return_dict['cat_cardinalities'] = [int(c) for c in cats]
#     return_dict['n_num_feature'] = 1

#     mimic_dataset = MimicDataset(input_dir, mode='train')
#     mimic_val_dataset = MimicDataset(input_dir, mode='val')

#     return_dict['dataset'] = mimic_dataset
#     return_dict['val_dataset'] = mimic_val_dataset

#     return return_dict

def load_synthetic_data(input_dir='./demo_data/synthetic_ehr', n_sample=None):
    '''
    Load the generated synthetic EHRs by PromptEHR.
    '''
    if input_dir is None:
        input_dir = './demo_data/synthetic_ehr'

    if not os.path.exists(input_dir):
        os.makedirs(input_dir)
        url = constants.SYNTHETIC_DATA_URL
        filename = wget.download(url, out=input_dir)
        print(f'Download synthetic EHRs to {input_dir}.')
    
    with open(os.path.join(input_dir,'data.pkl'), 'rb') as f:
        x = dill.load(f)

    if n_sample is not None:
        # cut to get smaller demo data
        x['visit'] = x['visit'][:n_sample]
        x['y'] = x['y'][:n_sample]
        x['feature'] = x['feature'][:n_sample]

    return x

